#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2020-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""Console script for MBoot module aka BLHost."""

import inspect
import json
import logging
import os
import shlex
import sys
from typing import Optional

import click

from spsdk import SPSDKError
from spsdk.apps.blhost_helper import (
    PROPERTIES_OVERRIDE,
    OemGenMasterShareHelp,
    OemSetMasterShareHelp,
    parse_key_prov_key_type,
    parse_property_tag,
    parse_trust_prov_key_type,
    parse_trust_prov_oem_key_type,
    progress_bar,
)
from spsdk.apps.utils import spsdk_logger
from spsdk.apps.utils.common_cli_options import (
    CommandsTreeGroup,
    isp_interfaces,
    spsdk_apps_common_options,
)
from spsdk.apps.utils.utils import (
    INT,
    SPSDKAppError,
    catch_spsdk_error,
    format_raw_data,
    get_interface,
    get_key,
    parse_file_and_size,
    parse_hex_data,
)
from spsdk.mboot import GenerateKeyBlobSelect, McuBoot, StatusCode, parse_property_value
from spsdk.mboot.error_codes import stringify_status_code
from spsdk.utils.images import BinaryImage


@click.group(name="blhost", no_args_is_help=True, cls=CommandsTreeGroup)
@isp_interfaces(uart=True, usb=True, sdio=True, lpcusbsio=True, buspal=True)
@spsdk_apps_common_options
@click.option(
    "-s",
    "--silent",
    is_flag=True,
    help="Silent mode suppresses progress bar and status response",
)
@click.pass_context
def main(
    ctx: click.Context,
    port: str,
    usb: str,
    sdio: str,
    buspal: str,
    lpcusbsio: str,
    use_json: bool,
    log_level: int,
    timeout: int,
    silent: bool,
) -> int:
    """Utility for communication with the bootloader on target."""
    log_level = log_level or logging.WARNING
    spsdk_logger.install(level=log_level)

    # print help for get-property if property tag is 0 or 'list-properties'
    if ctx.invoked_subcommand == "get-property":
        args = sys.argv[1:]
        # running this via pytest changes the args to a single arg, in that case skip
        if len(args) > 1 and "get-property" in args:
            tag_str = args[args.index("get-property") + 1]
            if parse_property_tag(tag_str) == 0:
                click.echo(ctx.command.commands["get-property"].help)  # type: ignore
                ctx.exit(0)

    # if --help is provided anywhere on command line, skip interface lookup and display help message
    if "--help" not in sys.argv[1:]:
        ctx.obj = {
            "interface": get_interface(
                module="mboot",
                port=port,
                usb=usb,
                sdio=sdio,
                timeout=timeout,
                buspal=buspal,
                lpcusbsio=lpcusbsio,
            ),
            "use_json": use_json,
            "suppress_progress_bar": use_json or silent or log_level < logging.WARNING,
            "silent": silent,
        }
    return 0


@main.command()
@click.argument("command_file", type=click.Path(file_okay=True))
@click.pass_context
def batch(ctx: click.Context, command_file: str) -> None:
    """Invoke blhost commands defined in command file.

    Command file contains one blhost command per line.
    example: "read-memory 0 4096 memory.bin"
    example: "get-property 24 # read target version"

    Comment are supported. Everything after '#' is a comment (just like in Python/Shell)

    Note: This is an early experimental format, it may change at any time.

    \b
    COMMAND_FILE    - path to blhost command file
    """
    with open(command_file) as f:
        for line in f.readlines():
            tokes = shlex.split(line, comments=True)
            if len(tokes) < 1:
                continue

            command_name, *command_args = tokes
            ctx.params = {}
            assert isinstance(ctx.parent, click.Context)
            assert isinstance(ctx.parent.command, click.Group)
            cmd_obj = ctx.parent.command.commands.get(command_name)
            if not cmd_obj:
                raise SPSDKError(f"Unknown command: {command_name}")
            cmd_obj.parse_args(ctx, command_args)
            ctx.invoke(cmd_obj, **ctx.params)


@main.command()
@click.argument("address", type=INT(), required=True)
@click.argument("argument", type=INT(), required=True)
@click.pass_context
def call(ctx: click.Context, address: int, argument: int) -> None:
    """Invokes code at an address, passing an argument to it.

    \b
    ADDRESS     - function code address
    ARGUMENT    - argument for the function
    """
    with McuBoot(ctx.obj["interface"]) as mboot:
        mboot.call(address, argument)
        display_output([], mboot.status_code, ctx.obj["use_json"], ctx.obj["silent"])


@main.command()
@click.argument("memory_id", type=INT(), required=True)
@click.argument("address", type=INT(), required=True)
@click.pass_context
def configure_memory(ctx: click.Context, address: int, memory_id: int) -> None:
    """Sets a config at internal memory to memory with ID.

    The specified configuration block must have been previously written to memory using the write-memory command.

    \b
    MEMORY_ID   - id of memory
    ADDRESS     - starting address
    """
    with McuBoot(ctx.obj["interface"]) as mboot:
        mboot.configure_memory(address, memory_id)  # type: ignore
        display_output([], mboot.status_code, ctx.obj["use_json"], ctx.obj["silent"])


@main.command()
@click.argument("address", type=INT(), required=True)
@click.argument("data", type=INT(base=16), required=True)
@click.argument(
    "lock",
    metavar="[nolock/lock]",
    type=click.Choice(["nolock", "lock"]),
    default="nolock",
)
@click.option(
    "-v / ",
    "--verify/--no-verify",
    is_flag=True,
    default=True,
    help="Verify write operation (verify by default)",
)
@click.pass_context
def efuse_program_once(
    ctx: click.Context, address: int, data: int, lock: str, verify: bool
) -> None:
    """Writes data to a specific efuse word.

    Each efuse bit can only be programmed once.

    \b
    ADDRESS - address of OTP word, not the shadowed memory address.
    DATA    - hex digits without prefix '0x'.
    """
    if lock == "lock":
        address = address | (1 << 24)
    with McuBoot(ctx.obj["interface"]) as mboot:
        response = mboot.efuse_program_once(address, data, verify=verify)
        display_output([response], mboot.status_code, ctx.obj["use_json"], ctx.obj["silent"])


@main.command()
@click.argument("address", type=INT(), required=True)
@click.pass_context
def efuse_read_once(ctx: click.Context, address: int) -> None:
    """Returns the contents of a specific efuse word.

    \b
    ADDRESS - is the address of OTP word, not the shadowed memory address.
    """
    with McuBoot(ctx.obj["interface"]) as mboot:
        response = mboot.efuse_read_once(address)
        display_output(
            None if response is None else [4, response],
            mboot.status_code,
            ctx.obj["use_json"],
            ctx.obj["silent"],
        )


@main.command()
@click.argument("address", type=INT(), required=True)
@click.argument("argument", type=INT(), required=True)
@click.argument("stackpointer", type=INT(), required=True)
@click.pass_context
def execute(ctx: click.Context, address: int, argument: int, stackpointer: int) -> None:
    """Jumps to code at the provided address.

    The system is returned to a reset state before the jump.
    The function <ARGUMENT> parameter is passed in R0 to the called code.

    The main stack pointer and process stack pointer registers are set to the <STACKPOINTER> parameter.
    If set to zero, the code being called should set the stack pointer before using the stack.

    \b
    ADDRESS      - Address of the application to run
    ARGUMENT     - Argument passed to the application
    STACKPOINTER - Stack pointer for the application
    """
    with McuBoot(ctx.obj["interface"]) as mboot:
        mboot.execute(address, argument, stackpointer)
        display_output([], mboot.status_code, ctx.obj["use_json"], ctx.obj["silent"])


@main.command()
@click.argument("address", type=INT(), required=True)
@click.argument("byte_count", type=INT(), required=True)
@click.argument("memory_id", type=INT(), required=False, default="0")
@click.pass_context
def flash_erase_region(ctx: click.Context, address: int, byte_count: int, memory_id: int) -> None:
    """Erases one or more sectors of the flash memory.

    The start <ADDRESS> and <BYTE_COUNT> must be a multiple of the word size.
    The entire sector(s) containing the start and end address is erased.

    \b
    ADDRESS     - starting address
    BYTE_COUNT  - number of bytes to erase
    MEMORY_ID   - id of memory to erase (default: 0)
    """
    with McuBoot(ctx.obj["interface"]) as mboot:
        mboot.flash_erase_region(address, byte_count, memory_id)
        display_output([], mboot.status_code, ctx.obj["use_json"], ctx.obj["silent"])


@main.command()
@click.argument("memory_id", type=INT(), required=False, default="0")
@click.pass_context
def flash_erase_all(ctx: click.Context, memory_id: int) -> None:
    """Performs an erase of the entire flash memory.

    \b
    MEMORY_ID   - id of memory to erase (default: 0)

    \b
    Note: excluding protected regions.
    """
    with McuBoot(ctx.obj["interface"]) as mboot:
        mboot.flash_erase_all(memory_id)
        display_output([], mboot.status_code, ctx.obj["use_json"], ctx.obj["silent"])


@main.command()
@click.pass_context
def flash_erase_all_unsecure(ctx: click.Context) -> None:
    """Erase complete flash memory and recover flash security section."""
    with McuBoot(ctx.obj["interface"]) as mboot:
        mboot.flash_erase_all_unsecure()
        display_output([], mboot.status_code, ctx.obj["use_json"], ctx.obj["silent"])


@main.command()
@click.argument("image_file_path", metavar="FILE", type=str, required=True)
@click.argument("erase", type=str, required=False, default="none")
@click.argument("memory_id", type=INT(), required=False, default="0")
@click.pass_context
def flash_image(ctx: click.Context, image_file_path: str, erase: str, memory_id: int) -> None:
    """Write the formatted image in <FILE> to the memory specified by memoryID.

    \b
    FILE       - path to image file
    ERASE      - string 'erase' determines if flash is erased before writing
    MEMORY_ID  - id of memory to erase (default: 0)
    """
    ALIGNMENT = 1024

    if not os.path.isfile(image_file_path):
        raise SPSDKError("The image file does not exist")
    mem_id = 0
    if erase not in ["erase", "none"]:
        try:
            mem_id = int(erase, 0)
        except ValueError as e:
            raise SPSDKError(
                "The option for erasing was not declared properly. Choose from 'erase' or 'none'."
            ) from e
    if memory_id:
        mem_id = memory_id
    bin_image = BinaryImage.load_binary_image(image_file_path)
    with McuBoot(ctx.obj["interface"]) as mboot:
        if erase == "erase":
            for segment in bin_image.sub_images:
                mboot.flash_erase_region(
                    address=segment.aligned_start(ALIGNMENT),
                    length=segment.aligned_length(ALIGNMENT),
                    mem_id=mem_id,
                )
                if mboot.status_code != StatusCode.SUCCESS:
                    display_output([], mboot.status_code, ctx.obj["use_json"], ctx.obj["silent"])
                    raise SPSDKAppError()
        for i, segment in enumerate(bin_image.sub_images, start=1):
            with progress_bar(
                suppress=ctx.obj["suppress_progress_bar"], label=f"Writing segment #{i}"
            ) as progress_callback:
                mboot.write_memory(
                    address=segment.absolute_address,
                    data=segment.export(),
                    mem_id=mem_id,
                    progress_callback=progress_callback,
                )
            display_output([], mboot.status_code, ctx.obj["use_json"], ctx.obj["silent"])


@main.command()
@click.argument("index", type=INT(), required=True)
@click.argument("byte_count", type=click.Choice(["4", "8"]), required=True)
@click.argument("data", type=INT(base=16), required=True)
@click.argument(
    "endianness",
    metavar="[LSB|MSB]",
    type=click.Choice(["LSB", "MSB"]),
    default="LSB",
    required=False,
)
@click.pass_context
def flash_program_once(
    ctx: click.Context, index: int, byte_count: str, data: int, endianness: str
) -> None:
    """Writes provided data to a specific program once field.

    \b
    INDEX       - fuse word index
    BYTE_COUNT  - width in bits (acceptable only 4 or 8-byte long data)
    DATA        - 4 or 8-byte-hex according to <byte_count>
    ENDIANNESS   - output sequence is specified by LSB (Default) or MSB
    """
    byte_order = "big" if endianness == "MSB" else "little"
    input_data = data.to_bytes(int(byte_count), byteorder=byte_order)  # type: ignore[arg-type]
    with McuBoot(ctx.obj["interface"]) as mboot:
        mboot.flash_program_once(index=index, data=input_data)
        display_output([], mboot.status_code, ctx.obj["use_json"], ctx.obj["silent"])


@main.command()
@click.argument("index", type=INT(), required=True)
@click.argument("byte_count", type=click.Choice(["4", "8"]), required=True, default=4)
@click.pass_context
def flash_read_once(ctx: click.Context, index: int, byte_count: str) -> None:
    """Returns the contents of a specific program once field.

    \b
    INDEX        - fuse word index
    BYTE_COUNT   - width in bits (acceptable only 4 or 8-byte long data)
    """
    with McuBoot(ctx.obj["interface"]) as mboot:
        response = mboot.flash_read_once(index=index, count=int(byte_count))
        display_output(
            None if response is None else [len(response), int.from_bytes(response, "little")],
            mboot.status_code,
            ctx.obj["use_json"],
            ctx.obj["silent"],
        )


@main.command()
@click.argument("key", type=str, required=True)
@click.pass_context
def flash_security_disable(ctx: click.Context, key: str) -> None:
    """Disable flash security by using of backdoor key.

    \b
    KEY        - key value as hex-string (8 bytes long)
    """
    if len(key) != 16:
        raise SPSDKError("Key length must be 8 bytes")
    try:
        key_bytes = bytes.fromhex(key)
    except ValueError as e:
        raise SPSDKError("Key is not a valid hex-string [A-Fa-f0-9]") from e
    with McuBoot(ctx.obj["interface"]) as mboot:
        mboot.flash_security_disable(backdoor_key=key_bytes)
        display_output([], mboot.status_code, ctx.obj["use_json"], ctx.obj["silent"])


@main.command()
@click.argument("address", type=INT(), required=True)
@click.argument("length", type=INT(), required=True)
@click.argument("option", type=click.Choice(["0", "1"]), required=True)
@click.argument("out_file", metavar="FILE", type=click.File("wb"), required=False)
@click.option("-h", "--use-hexdump", is_flag=True, default=False, help="Use hexdump format")
@click.pass_context
def flash_read_resource(
    ctx: click.Context,
    address: int,
    length: int,
    option: str,
    out_file: click.File,
    use_hexdump: bool,
) -> None:
    """Read resource of flash module.

    Reads the contents of Flash IFR or Flash Firmware ID as specified by option and
    writes result to file or stdout if file is not specified.

    \b
    ADDRESS      - Start address
    LENGTH       - Number of bytes to read. Must be 4-byte aligned.
    OPTION       - Area to be read. 0 means Flash IFR, 1 means Flash Firmware ID.
    OUT_FILE     - Path to file, where the output will be stored
    """
    with McuBoot(ctx.obj["interface"]) as mboot:
        response = mboot.flash_read_resource(address=address, length=length, option=int(option))

        if response:
            if out_file:
                out_file.write(response)  # type: ignore
            else:
                click.echo(format_raw_data(response, use_hexdump=use_hexdump))

        display_output(
            [len(response) if response else 0],
            mboot.status_code,
            ctx.obj["use_json"],
            ctx.obj["silent"],
            f"Read {len(response) if response else 0} of {length} bytes.",
        )


@main.command()
@click.argument("address", type=INT(), required=True)
@click.argument("byte_count", type=INT(), required=True)
@click.argument("pattern", type=INT(), required=True)
@click.argument(
    "pattern_format",
    metavar="format",
    type=click.Choice(["word", "short", "byte"]),
    required=False,
    default="word",
)
@click.pass_context
def fill_memory(
    ctx: click.Context, address: int, byte_count: int, pattern: int, pattern_format: str
) -> None:
    """Fills the memory with a pattern.

    \b
    ADDRESS     - starting address
    BYTE_COUNT  - number of bytes to fill
    PATTERN     - pattern to fill
    FORMAT      - format of the pattern [word|short|byte] (default: word)
    """
    del pattern_format  # temporary workaround for not unused parameter
    with McuBoot(ctx.obj["interface"]) as mboot:
        mboot.fill_memory(address, byte_count, pattern)
        display_output([], mboot.status_code, ctx.obj["use_json"], ctx.obj["silent"])


@main.command()
@click.argument("address", type=INT(), required=True)
@click.argument("data_source", metavar="FILE[,BYTE_COUNT] | {{HEX-DATA}}", type=str, required=True)
@click.argument("memory_id", type=INT(), required=False, default="0")
@click.pass_context
def fuse_program(ctx: click.Context, address: int, data_source: str, memory_id: int) -> None:
    """Program fuse.

    \b
    ADDRESS     - starting address
    FILE        - write the content of this file
    BYTE_COUNT  - if specified, load only first BYTE_COUNT number of bytes from file
    HEX-DATA    - string of hex values: {{112233}}, {{11 22 33}}
                - when using Jupyter notebook, use [[ ]] instead of {{ }}: eg. [[11 22 33]]
    MEMORY_ID   - id of memory to read from (default: 0)
    """
    try:
        data = parse_hex_data(data_source)
    except SPSDKError:
        file_path, size = parse_file_and_size(data_source)
        with open(file_path, "rb") as f:
            data = f.read(size)

    with McuBoot(ctx.obj["interface"]) as mboot:
        response = mboot.fuse_program(address, data, memory_id)
        display_output(
            [len(data)] if response else None,
            mboot.status_code,
            ctx.obj["use_json"],
            ctx.obj["silent"],
        )


@main.command()
@click.argument("address", type=INT(), required=True)
@click.argument("byte_count", type=INT(), required=True)
@click.argument("out_file", metavar="FILE", type=click.File("wb"), required=False)
@click.argument("memory_id", type=INT(), default="0", required=False)
@click.option("-h", "--use-hexdump", is_flag=True, default=False, help="Use hexdump format")
@click.pass_context
def fuse_read(
    ctx: click.Context,
    address: int,
    byte_count: int,
    out_file: click.File,
    memory_id: int,
    use_hexdump: bool,
) -> None:
    """Reads the fuse and writes it to the file or stdout.

    Returns the contents of memory at the given <ADDRESS>, for a specified <BYTE_COUNT>.

    \b
    ADDRESS     - starting address
    BYTE_COUNT  - number of bytes to read
    FILE        - store result into this file, if not specified use stdout
    MEMORY_ID   - id of memory to read from (default: 0)
    """
    with McuBoot(ctx.obj["interface"]) as mboot:
        response = mboot.fuse_read(address, byte_count, memory_id)

    if response:
        if out_file:
            out_file.write(response)  # type: ignore
        else:
            click.echo(format_raw_data(response, use_hexdump=use_hexdump))

    display_output(
        [len(response) if response else 0],
        mboot.status_code,
        ctx.obj["use_json"],
        ctx.obj["silent"],
        f"Read {len(response) if response else 0} of {byte_count} bytes.",
    )


@main.command()
@click.pass_context
def list_memory(ctx: click.Context) -> None:
    """Lists all memories, supported by the current device."""
    with McuBoot(ctx.obj["interface"]) as mboot:
        print("Internal Flash:")
        int_flash = mboot._get_internal_flash()  # pylint: disable=protected-access
        for flash in int_flash:
            print(f"    {flash}")
        print("Internal RAM:")
        int_ram = mboot._get_internal_ram()  # pylint: disable=protected-access
        for ram in int_ram:
            print(f"    {ram}")
        print("External Memories:")
        ext_mem = mboot._get_ext_memories()  # pylint: disable=protected-access
        for mem in ext_mem:
            print(f"{(mem.name)}:\n  {mem}")


@main.command()
@click.argument("boot_file", metavar="FILE", type=click.File("rb"))
@click.pass_context
def load_image(ctx: click.Context, boot_file: click.File) -> None:
    """Sends a boot image file to the device.

    Only binary file is supported.
    The <FILE> must be a bootable image which contains the boot image header supported by MCU bootloader.

    \b
    FILE  - boot file to load
    """
    data = boot_file.read()  # type: ignore
    with McuBoot(ctx.obj["interface"]) as mboot:
        with progress_bar(
            suppress=ctx.obj["suppress_progress_bar"], label="Loading image"
        ) as progress_callback:
            mboot.load_image(data, progress_callback)
        display_output([], mboot.status_code, ctx.obj["use_json"], ctx.obj["silent"])


@main.command()
@click.argument("property_tag", type=str, required=True)
@click.argument("index", type=INT(), default="0")
@click.option(
    "-f",
    "--family",
    type=click.Choice(list(PROPERTIES_OVERRIDE.keys())),
    help="Provide family. So far supported 'kw45xx' and 'k32w1xx'",
)
@click.pass_context
def get_property(
    ctx: click.Context,
    property_tag: str,
    index: int,
    family: str,
) -> None:
    """Queries various bootloader properties and settings.

    Each supported property has a unique <PROPERTY_TAG>.

    \b
    PROPERTY_TAG - number or name representing the requested property
    INDEX        - id/index of the memory (default: 0)

    \b
    Available properties:
     0 or 'list-properties'             List all properties
     1 or 'current-version'             Bootloader version
     2 or 'available-peripherals'       Available peripherals
     3 or 'flash-start-address'         Start of program flash, <index> is required
     4 or 'flash-size-in-bytes'         Size of program flash, <index> is required
     5 or 'flash-sector-size'           Size of flash sector, <index> is required
     6 or 'flash-block-count'           Blocks in flash array, <index> is required
     7 or 'available-commands'          Available commands
     8 or 'check-status'                Check Status, <status id> is required
     9 or 'reserved'
    10 or 'verify-writes'               Verify Writes flag
    11 or 'max-packet-size'             Max supported packet size
    12 or 'reserved-regions'            Reserved regions
    13 or 'reserved'
    14 or 'ram-start-address'           Start of RAM, <index> is required
    15 or 'ram-size-in-bytes'           Size of RAM, <index> is required
    16 or 'system-device-id'            System device identification
    17 or 'security-state'              Flash security state
    18 or 'unique-device-id'            Unique device identification
    19 or 'flash-fac-support'           FAC support flag
    20 or 'flash-access-segment-size'   FAC segment size
    21 or 'flash-access-segment-count'  FAC segment count
    22 or 'flash-read-margin'           Read margin level of program flash
    23 or 'qspi/otfad-init-status'      QuadSpi initialization status
    24 or 'target-version'              Target version
    25 or 'external-memory-attributes'  External memory attributes, <memoryId> is required
    26 or 'reliable-update-status'      Reliable update status
    27 or 'flash-page-size'             Flash page size, <index> is required
    28 or 'irq-notify-pin'              Interrupt notifier pin
    29 or 'pfr-keystore_update-opt'     PFR key store update option
    30 or 'byte-write-timeout-ms'       Byte write timeout in ms
    31 or 'fuse-locked-status'          Fuse Locked Status
    for kw45xx/k32w1xx devices:
    10 or 'verify-erases'               Verify Erases flag
    20 or 'boot status'                 Value of Boot Status Register
    21 or 'loadable-fw-version'         LoadableFWVersion
    22 or 'fuse-program-voltage'        Fuse Program Voltage

    \b
    Note: Not all the properties are available for all devices.
    """
    property_tag_int = parse_property_tag(property_tag, family)
    with McuBoot(ctx.obj["interface"]) as mboot:
        response = mboot.get_property(property_tag_int, index=index)  # type: ignore
        property_text = (
            str(parse_property_value(property_tag_int, response, None, family))
            if response
            else None
        )
        display_output(
            response, mboot.status_code, ctx.obj["use_json"], ctx.obj["silent"], property_text
        )


@main.command()
@click.argument("property_tag", type=str, required=True)
@click.argument("value", type=INT(), required=True)
@click.option(
    "-f",
    "--family",
    type=click.Choice(list(PROPERTIES_OVERRIDE.keys())),
    help="Provide family. So far supported 'kw45xx' and 'k32w1xx'",
)
@click.pass_context
def set_property(
    ctx: click.Context,
    property_tag: str,
    value: int,
    family: str,
) -> None:
    """Changes properties and options in the bootloader.

    Accepts the same <PROPERTY_TAG> used with the get-property sub-command;
    however, only some properties are writable.

    \b
    PROPERTY_TAG - number or name representing the requested property
    VALUE        - value to set

    \b
    Available properties to set:
    10 or 'verify-writes'               Verify Writes flag
    22 or 'flash-read-margin'           Read margin level of program flash
    28 or 'irq-notify-pin'              Interrupt notifier pin
    29 or 'pfr-keystore_update-opt'     PFR key store update option
    30 or 'byte-write-timeout-ms'       Byte write timeout in ms
    for kw45xx/k32w1xx devices:
    10 or 'verify-erases'               Verify Erases flag
    22 or 'fuse-program-voltage'        Fuse Program Voltage

    \b
    Note: Not all properties can be set on all devices.
    """
    property_tag_int = parse_property_tag(property_tag, family)
    with McuBoot(ctx.obj["interface"]) as mboot:
        mboot.set_property(prop_tag=property_tag_int, value=value)  # type: ignore
        display_output([], mboot.status_code, ctx.obj["use_json"], ctx.obj["silent"])


@main.command()
@click.argument("address", type=INT(), required=True)
@click.argument("byte_count", type=INT(), required=True)
@click.argument("out_file", metavar="FILE", type=click.File("wb"), required=False)
@click.argument("memory_id", type=INT(), default="0", required=False)
@click.option("-h", "--use-hexdump", is_flag=True, default=False, help="Use hexdump format")
@click.option(
    "-f",
    "--fast-mode",
    is_flag=True,
    default=False,
    help="Fast mode for USB-HID data transfer, not reliable !!!",
)
@click.pass_context
def read_memory(
    ctx: click.Context,
    address: int,
    byte_count: int,
    out_file: click.File,
    memory_id: int,
    use_hexdump: bool,
    fast_mode: bool,
) -> None:
    """Reads the memory and writes it to the file or stdout.

    Returns the contents of memory at the given <ADDRESS>, for a specified <BYTE_COUNT>.

    \b
    ADDRESS     - starting address
    BYTE_COUNT  - number of bytes to read
    FILE        - store result into this file, if not specified use stdout.
                 If needed specify MEMORY_ID argument, use '-' instead name of file
                 to print to stdout.
    MEMORY_ID   - id of memory to read from (default: 0)
    """
    with McuBoot(ctx.obj["interface"]) as mboot:
        with progress_bar(
            suppress=ctx.obj["suppress_progress_bar"], label="Reading memory"
        ) as progress_callback:
            response = mboot.read_memory(
                address, byte_count, memory_id, progress_callback, fast_mode
            )

    if response:
        if out_file and out_file.name != "<stdout>":
            out_file.write(response)  # type: ignore
        else:
            click.echo(format_raw_data(response, use_hexdump=use_hexdump))

    display_output(
        [len(response) if response else 0],
        mboot.status_code,
        ctx.obj["use_json"],
        ctx.obj["silent"],
        f"Read {len(response) if response else 0} of {byte_count} bytes.",
    )


@main.command()
@click.argument("sb_file", metavar="FILE", type=click.File("rb"), required=True)
@click.option(
    "-c",
    "--check-errors",
    is_flag=True,
    default=False,
    help=(
        "This flag should be used when the `receive-sb-file` operation fails using USB interface. "
        "Without this flag USB transfer is significantly faster (roughly 20x) "
        "However, the status code might be misleading in case of an error. "
        "In case of an error using USB interface, "
        "rerun `receive-sb-file` with this setting for clearer error message. "
        "This setting has no effect interfaces other than USB."
    ),
)
@click.pass_context
def receive_sb_file(ctx: click.Context, sb_file: click.File, check_errors: bool) -> None:
    """Receives a file in a Secure Binary (SB) format.

    An SB file is an encapsulated, binary stream of bootloader commands that can be optionally encrypted.

    \b
    FILE    - SB file to send to the target
    """
    with McuBoot(ctx.obj["interface"]) as mboot:
        with progress_bar(
            suppress=ctx.obj["suppress_progress_bar"], label="Sending SB file"
        ) as progress_callback:
            data = sb_file.read()  # type: ignore
            mboot.receive_sb_file(data, progress_callback, check_errors)
        display_output([], mboot.status_code, ctx.obj["use_json"], ctx.obj["silent"])


@main.command()
@click.argument("address", type=INT(), required=True)
@click.pass_context
def reliable_update(ctx: click.Context, address: int) -> None:
    """Reliable Update.

    \b
    ADDRESS     - starting address
    """
    with McuBoot(ctx.obj["interface"]) as mboot:
        mboot.reliable_update(address)
        display_output([], mboot.status_code, ctx.obj["use_json"], ctx.obj["silent"])


@main.command()
@click.pass_context
def reset(ctx: click.Context) -> None:
    """Resets the device.

    A response packet is sent before resetting the device.
    """
    with McuBoot(ctx.obj["interface"]) as mboot:
        mboot.reset(reopen=False)
    display_output([], mboot.status_code, ctx.obj["use_json"], ctx.obj["silent"])


@main.command()
@click.argument("address", type=INT(), required=True)
@click.argument("data_source", metavar="FILE[,BYTE_COUNT] | {{HEX-DATA}}", type=str, required=True)
@click.argument("memory_id", type=INT(), required=False, default="0")
@click.pass_context
def write_memory(ctx: click.Context, address: int, data_source: str, memory_id: int) -> None:
    """Writes memory from a file or a hex-data.

    Writes memory specified by <MEMORY_ID> at <ADDRESS> from <FILE> or <HEX-DATA>
    Writes a provided buffer to a specified <BYTE_COUNT> in memory.

    \b
    ADDRESS     - starting address
    FILE        - write the content of this file
    BYTE_COUNT  - if specified, load only first BYTE_COUNT number of bytes from file
    HEX-DATA    - string of hex values: {{112233}}, {{11 22 33}}
                - when using Jupyter notebook, use [[ ]] instead of {{ }}: eg. [[11 22 33]]
    MEMORY_ID   - id of memory to read from (default: 0)
    """
    try:
        data = parse_hex_data(data_source)
    except SPSDKError:
        file_path, size = parse_file_and_size(data_source)
        with open(file_path, "rb") as f:
            data = f.read(size)

    with McuBoot(ctx.obj["interface"]) as mboot:
        with progress_bar(
            suppress=ctx.obj["suppress_progress_bar"], label="Writing memory"
        ) as progress_callback:
            response = mboot.write_memory(address, data, memory_id, progress_callback)
        display_output(
            [len(data)] if response else None,
            mboot.status_code,
            ctx.obj["use_json"],
            ctx.obj["silent"],
        )


@main.command()
@click.argument("dek_file", type=click.File("rb"), required=True)
@click.argument("blob_file", type=click.File("wb"), required=True)
@click.argument(
    "key_sel",
    metavar="[KEY_SEL]",
    type=click.Choice(["0", "1", "2", "3", "OPTMK", "ZMK", "CMK"]),
    default="0",
)
@click.pass_context
def generate_key_blob(
    ctx: click.Context, dek_file: click.File, blob_file: click.File, key_sel: str
) -> None:
    """Generates the Key Blob, and writes it to the file.

    <KEY_SEL> selects the blob key encryption key(BKEK) used to generate the key blob.

    \b
    DEK_FILE     - the file with the binary DEK key
    BLOB_FILE    - the generated file with a binary key blob
    KEY_SEL      - select the BKEK used to wrap the BK and generate the blob.
                   For devices with SNVS, valid options of [key_sel] are
                        0, 1 or OTPMK: OTPMK from FUSE or OTP(default),
                        2 or ZMK: ZMK from SNVS,
                        3 or CMK: CMK from SNVS,
                   For devices without SNVS, this option will be ignored.
    """
    with McuBoot(ctx.obj["interface"]) as mboot:
        data = dek_file.read()  # type: ignore
        key_sel_int = int(key_sel) if key_sel.isnumeric() else GenerateKeyBlobSelect.get(key_sel)
        assert isinstance(key_sel_int, int)
        write_response = mboot.generate_key_blob(data, key_sel=key_sel_int)
        if write_response:
            blob_file.write(write_response)  # type: ignore
        display_output(
            [mboot.status_code, len(write_response)] if write_response else None,
            mboot.status_code,
            ctx.obj["use_json"],
            ctx.obj["silent"],
        )


@main.group()
@click.pass_context
def key_provisioning(ctx: click.Context) -> None:  # pylint: disable=unused-argument
    """Group of sub-commands related to key provisioning."""


@key_provisioning.command()
@click.pass_context
def enroll(ctx: click.Context) -> None:
    """Enrolls key provisioning feature. No argument for this operation."""
    with McuBoot(ctx.obj["interface"]) as mboot:
        mboot.kp_enroll()
        display_output([], mboot.status_code, ctx.obj["use_json"], ctx.obj["silent"])


@main.command()
@click.argument("file", metavar="FILE", type=click.File("rb"), required=True)
@click.pass_context
def program_aeskey(ctx: click.Context, file: click.File) -> None:
    """Sends raw binary, which contains an aes key, to the devices and program it to the OTP field.

    \b
    FILE    - file, which contains an aes key
    """
    data = file.read()  # type: ignore
    with McuBoot(ctx.obj["interface"]) as mboot:
        mboot.write_memory(address=0x0, data=data, mem_id=0x200)
        display_output([], mboot.status_code, ctx.obj["use_json"], ctx.obj["silent"])


@key_provisioning.command(name="set_user_key")
@click.option(
    "-s",
    "--key-size",
    type=INT(),
    required=False,
    help=(
        "Key size in bits. If this field is defined,"
        " the command could load as text as binary form of key."
    ),
)
@click.argument("key_type", metavar="TYPE", type=str, required=True)
@click.argument("file_and_size", metavar="FILE[,SIZE]", type=str, required=True)
@click.pass_context
def set_user_key(ctx: click.Context, key_type: str, file_and_size: str, key_size: int) -> None:
    """Sends the user key specified by type to the bootloader.

    <FILE> is the binary file containing user key plain text.
    If <SIZE> is not specified, the entire <FILE> will be sent.
    Otherwise, blhost only sends the first <SIZE> bytes.

    \b
    If the '--key-size' option is defined, the argument 'file_and_size' accepts those formats:
     - filename to binary file
     - filename to text file
     - string with key
    The size part is ignored in this case.

    \b
    TYPE  - Type of user key
    FILE  - Binary file containing user key plaintext
    SIZE  - If not specified, the entire <file> will be sent. Otherwise, only send
            the first <size> bytes.

    \b
    Available KEY TYPES:
     2 or 'OTFADKEK'    OTFAD key
     3 or 'SBKEK'       SB file encryption key
     7 or 'PRINCE0'     Prince region 0 encryption key
     8 or 'PRINCE1'     Prince region 1 encryption key
     9 or 'PRINCE2'     Prince region 2 encryption key
    11 or 'USERKEK'     User/Boot-image encryption key
    12 or 'UDS'         Universal Device Secret for DICE

    \b
    Note: The valid options of <type> and corresponding <size> are documented
          in the target's Reference Manual or User Manual.
    Note: Names are case insensitive
    """
    file_path, size = parse_file_and_size(file_and_size)
    key_type_int = parse_key_prov_key_type(key_type)

    if not key_size:
        key_size = (size if size > 0 else os.path.getsize(file_path)) * 8

    key_data = get_key(file_path, expected_size=key_size // 8)

    with McuBoot(ctx.obj["interface"]) as mboot:
        mboot.kp_set_user_key(key_type=key_type_int, key_data=key_data)  # type: ignore
        display_output([], mboot.status_code, ctx.obj["use_json"], ctx.obj["silent"])


@key_provisioning.command(name="set_key")
@click.argument("key_type", metavar="TYPE", type=str, required=True)
@click.argument("key_size", metavar="SIZE", type=INT(), required=True)
@click.pass_context
def set_key(ctx: click.Context, key_type: str, key_size: int) -> None:
    """Generates a size bytes of the key specified by the type.

    \b
    TYPE  - type of key to generate,
    SIZE  - size of key to generate

    \b
    Available KEY TYPES:
     2 or 'OTFADKEK'    OTFAD key
     3 or 'SBKEK'       SB file encryption key
     7 or 'PRINCE0'     Prince region 0 encryption key
     8 or 'PRINCE1'     Prince region 1 encryption key
     9 or 'PRINCE2'     Prince region 2 encryption key
    11 or 'USERKEK'     User/Boot-image encryption key
    12 or 'UDS'         Universal Device Secret for DICE

    \b
    Note: The valid options of <type> and corresponding <size> are documented
          in the target's Reference Manual or User Manual.
    Note: Names are case insensitive
    """
    key_type_int = parse_key_prov_key_type(key_type)
    with McuBoot(ctx.obj["interface"]) as mboot:
        mboot.kp_set_intrinsic_key(key_type_int, key_size)
        display_output([], mboot.status_code, ctx.obj["use_json"], ctx.obj["silent"])


@key_provisioning.command(name="write_key_nonvolatile")
@click.argument("memory_id", metavar="memoryID", type=INT(), default="0")
@click.pass_context
def write_key_nonvolatile(ctx: click.Context, memory_id: int) -> None:
    """Writes the key to nonvolatile memory.

    \b
    memoryID  - ID of the non-volatile memory, default: 0
    """
    with McuBoot(ctx.obj["interface"]) as mboot:
        mboot.kp_write_nonvolatile(memory_id)
        display_output([], mboot.status_code, ctx.obj["use_json"], ctx.obj["silent"])


@key_provisioning.command(name="read_key_nonvolatile")
@click.argument("memory_id", metavar="memoryID", type=INT(), default="0")
@click.pass_context
def read_key_nonvolatile(ctx: click.Context, memory_id: int) -> None:
    """Loads the key from nonvolatile memory to bootloader.

    \b
    memoryID  - ID of the non-volatile memory, default: 0
    """
    with McuBoot(ctx.obj["interface"]) as mboot:
        mboot.kp_read_nonvolatile(memory_id)
        display_output([], mboot.status_code, ctx.obj["use_json"], ctx.obj["silent"])


@key_provisioning.command(name="write_key_store")
@click.argument("file_and_size", metavar="FILE[,SIZE]", type=str, required=True)
@click.pass_context
def write_key_store(ctx: click.Context, file_and_size: str) -> None:
    """Sends the key store to the bootloader.

    <FILE> is the binary file containing key store.
    If <SIZE> is not specified, the entire <FILE> will be sent.
    Otherwise, only send the first <SIZE> bytes.

    \b
    FILE  - Binary file containing key store.
    SIZE  - If not specified, the entire <file> will be sent. Otherwise, only send
            the first <size> bytes.
    """
    file_path, size = parse_file_and_size(file_and_size)

    with open(file_path, "rb") as key_file:
        key_data = key_file.read(size)

    with McuBoot(ctx.obj["interface"]) as mboot:
        mboot.kp_write_key_store(key_data)
        display_output([], mboot.status_code, ctx.obj["use_json"], ctx.obj["silent"])


@key_provisioning.command(name="read_key_store")
@click.argument("key_store_file", metavar="FILE", type=click.File("wb"), required=True)
@click.pass_context
def read_key_store(ctx: click.Context, key_store_file: click.File) -> None:
    """Reads the key store from the bootloader to host(PC).

    <FILE> is the binary file to store the key store.

    \b
    FILE  - Binary file to save the key store.
    """
    with McuBoot(ctx.obj["interface"]) as mboot:
        response = mboot.kp_read_key_store()
        if response:
            key_store_file.write(response)  # type: ignore
        display_output(
            [len(response)] if response else None,
            mboot.status_code,
            ctx.obj["use_json"],
            ctx.obj["silent"],
        )


@main.group()
@click.pass_context
def trust_provisioning(ctx: click.Context) -> None:  # pylint: disable=unused-argument
    """Group of sub-commands related to trust provisioning."""


@trust_provisioning.command(name="hsm_store_key")
@click.argument("key_type", metavar="KEY_TYPE", type=str, required=True)
@click.argument("key_property", metavar="KEY_PROPERTY", type=INT(), required=True)
@click.argument("key_input_addr", metavar="KEY_INPUT_ADDR", type=INT(), required=True)
@click.argument("key_input_size", metavar="KEY_INPUT_SIZE", type=INT(), required=True)
@click.argument("key_blob_output_addr", metavar="KEY_BLOB_OUTPUT_ADDR", type=INT(), required=True)
@click.argument("key_blob_output_size", metavar="KEY_BLOB_OUTPUT_SIZE", type=INT(), required=True)
@click.pass_context
def hsm_store_key(
    ctx: click.Context,
    key_type: str,
    key_property: int,
    key_input_addr: int,
    key_input_size: int,
    key_blob_output_addr: int,
    key_blob_output_size: int,
) -> None:
    """Stores known keys, and generate the corresponding key blob.

    It wraps the known key, which is given by the customer,
    using NXP_CUST_KEK_EXT_SK, and output the RFC3396 key blob.

    \b
    KEY_TYPE              - Type of key to generate (CKDFK, HKDFK, HMACK, CMACK, AESK, KUOK)
    KEY_PROPERTY          - Bit 0: Key Size, 0 for 128bit, 1 for 256bit. Bits 30-31: set key protection CSS mode
    KEY_INPUT_ADDR        - The input buffer address where the key locates at
    KEY_INPUT_SIZE        - The byte count of the key
    KEY_BLOB_OUTPUT_ADDR  - The output buffer address where ROM writes the key blob to
    KEY_BLOB_OUTPUT_SIZE  - The output buffer size in byte
    """
    key_type_int = parse_trust_prov_key_type(key_type)
    with McuBoot(ctx.obj["interface"]) as mboot:
        response = mboot.tp_hsm_store_key(
            key_type_int,
            key_property,
            key_input_addr,
            key_input_size,
            key_blob_output_addr,
            key_blob_output_size,
        )

        extra_output = ""
        if response:
            key_header = response[0]
            key_blob_size = response[1]
            if mboot.status_code == StatusCode.SUCCESS:
                extra_output = "Output data size/value(s) is(are):\n"
                extra_output += (
                    f"\tKey Header: {key_header} ({hex(key_header)})\n"
                    f"\tKey Blob size: {key_blob_size} ({hex(key_blob_size)})"
                )
        display_output(
            response, mboot.status_code, ctx.obj["use_json"], ctx.obj["silent"], extra_output
        )


@trust_provisioning.command(name="hsm_gen_key")
@click.argument("key_type", metavar="KEY_TYPE", type=str, required=True)
@click.argument("reserved", metavar="RESERVED", type=INT(), required=True)
@click.argument("key_blob_output_addr", metavar="KEY_BLOB_OUTPUT_ADDR", type=INT(), required=True)
@click.argument("key_blob_output_size", metavar="KEY_BLOB_OUTPUT_SIZE", type=INT(), required=True)
@click.argument("ecdsa_puk_output_addr", metavar="ECDSA_PUK_OUTPUT_ADDR", type=INT(), required=True)
@click.argument("ecdsa_puk_output_size", metavar="ECDSA_PUK_OUTPUT_SIZE", type=INT(), required=True)
@click.pass_context
def hsm_gen_key(
    ctx: click.Context,
    key_type: str,
    reserved: int,
    key_blob_output_addr: int,
    key_blob_output_size: int,
    ecdsa_puk_output_addr: int,
    ecdsa_puk_output_size: int,
) -> None:
    """Creates OEM common keys, including encryption keys and signing keys.

    It outputs the key blob, which is wrapped by NXP_CUST_KEK_IN_SK
    and the public portion of the signing key.

    \b
    KEY_TYPE              - Type of key to generate (MFWISK, MFWENCK, GENSIGNK, GETCUSTMKSK)
    RESERVED              - Reserved must be 0
    KEY_BLOB_OUTPUT_ADDR  - Output buffer address where ROM writes the key blob to
    KEY_BLOB_OUTPUT_SIZE  - Output buffer size in bytes
    ECDSA_PUK_OUTPUT_ADDR - Output buffer address where ROM writes the public key to
    ECDSA_PUK_OUTPUT_SIZE - Output buffer size in bytes
    """
    key_type_int = parse_trust_prov_oem_key_type(key_type)
    with McuBoot(ctx.obj["interface"]) as mboot:
        response = mboot.tp_hsm_gen_key(
            key_type_int,
            reserved,
            key_blob_output_addr,
            key_blob_output_size,
            ecdsa_puk_output_addr,
            ecdsa_puk_output_size,
        )

        extra_output = ""
        if response:
            keyblob_size = response[0]
            ecdsa_puk_size = response[1]
            if mboot.status_code == StatusCode.SUCCESS:
                extra_output = "Output data size/value(s) is(are):\n"
            else:
                extra_output = (
                    "Output buffer(s) is(are) smaller than the minimum requested which is(are):\n"
                )
            extra_output += (
                f"\tKey Blob size: {keyblob_size} ({hex(keyblob_size)})\n"
                f"\tECDSA Puk size: {ecdsa_puk_size} ({hex(ecdsa_puk_size)})"
            )
        display_output(
            response, mboot.status_code, ctx.obj["use_json"], ctx.obj["silent"], extra_output
        )


@trust_provisioning.command(name="hsm_enc_blk")
@click.argument(
    "mfg_cust_mk_sk_0_blob_input_addr",
    metavar="MFG_CUST_MK_SK_0_BLOB_INPUT_ADDR",
    type=INT(),
    required=True,
)
@click.argument(
    "mfg_cust_mk_sk_0_blob_input_size",
    metavar="MFG_CUST_MK_SK_0_BLOB_INPUT_SIZE",
    type=INT(),
    required=True,
)
@click.argument("kek_id", metavar="KEK_ID", type=str, required=True)
@click.argument("sb3_header_input_addr", metavar="SB3_HEADER_INPUT_ADDR", type=INT(), required=True)
@click.argument("sb3_header_input_size", metavar="SB3_HEADER_INPUT_SIZE", type=INT(), required=True)
@click.argument("block_num", metavar="BLOCK_NUM", type=INT(), required=True)
@click.argument("block_data_addr", metavar="BLOCK_DATA_ADDR", type=INT(), required=True)
@click.argument("block_data_size", metavar="BLOCK_DATA_SIZE", type=INT(), required=True)
@click.pass_context
def hsm_enc_blk(
    ctx: click.Context,
    mfg_cust_mk_sk_0_blob_input_addr: int,
    mfg_cust_mk_sk_0_blob_input_size: int,
    kek_id: str,
    sb3_header_input_addr: int,
    sb3_header_input_size: int,
    block_num: int,
    block_data_addr: int,
    block_data_size: int,
) -> None:
    """Encrypts the given SB3 data block.

    \b
    MFG_CUST_MK_SK_0_BLOB_INPUT_ADDR - The input buffer address where the CKDF Master Key Blob locates at
    MFG_CUST_MK_SK_0_BLOB_INPUT_SIZE - The byte count of the CKDF Master Key Blob
    KEK_ID                           - The CKDF Master Key Encryption Key ID
    (0x10: NXP_CUST_KEK_INT_SK, 0x11: NXP_CUST_KEK_EXT_SK)
    SB3_HEADER_INPUT_ADDR            - The input buffer address where the SB3 Header(block0) locates at
    SB3_HEADER_INPUT_SIZE            - The byte count of the SB3 Header
    BLOCK_NUM                        - The index of the block. Due to SB3 Header(block 0) is always unencrypted,
    the index starts from block1
    BLOCK_DATA_ADDR                  - The buffer address where the SB3 data block locates at
    BLOCK_DATA_SIZE                  - The byte count of the SB3 data block
    """
    kek_id_int = parse_trust_prov_key_type(kek_id)
    with McuBoot(ctx.obj["interface"]) as mboot:
        mboot.tp_hsm_enc_blk(
            mfg_cust_mk_sk_0_blob_input_addr,
            mfg_cust_mk_sk_0_blob_input_size,
            kek_id_int,
            sb3_header_input_addr,
            sb3_header_input_size,
            block_num,
            block_data_addr,
            block_data_size,
        )
        display_output([], mboot.status_code, ctx.obj["use_json"], ctx.obj["silent"])


@trust_provisioning.command(name="hsm_enc_sign")
@click.argument("key_blob_input_addr", metavar="KEY_BLOB_INPUT_ADDR", type=INT(), required=True)
@click.argument("key_blob_input_size", metavar="KEY_BLOB_INPUT_SIZE", type=INT(), required=True)
@click.argument("block_data_input_addr", metavar="BLOCK_DATA_INPUT_ADDR", type=INT(), required=True)
@click.argument("block_data_input_size", metavar="BLOCK_DATA_INPUT_SIZE", type=INT(), required=True)
@click.argument("signature_output_addr", metavar="SIGNATURE_OUTPUT_ADDR", type=INT(), required=True)
@click.argument("signature_output_size", metavar="SIGNATURE_OUTPUT_SIZE", type=INT(), required=True)
@click.pass_context
def hsm_enc_sign(
    ctx: click.Context,
    key_blob_input_addr: int,
    key_blob_input_size: int,
    block_data_input_addr: int,
    block_data_input_size: int,
    signature_output_addr: int,
    signature_output_size: int,
) -> None:
    """Signs the given data.

    It uses the private key in the given key blob, which is generated by HSM_GEN_KEY.

    \b
    KEY_BLOB_INPUT_ADDR   - The input buffer address where signing key blob locates at
    KEY_BLOB_INPUT_SIZE   - The byte count of the signing key blob
    BLOCK_DATA_INPUT_ADDR - The input buffer address where the data locates at
    BLOCK_DATA_INPUT_SIZE - The byte count of the data
    SIGNATURE_OUTPUT_ADDR - The output buffer address where ROM writes the signature to
    SIGNATURE_OUTPUT_SIZE - The output buffer size in byte
    """
    with McuBoot(ctx.obj["interface"]) as mboot:
        response = mboot.tp_hsm_enc_sign(
            key_blob_input_addr,
            key_blob_input_size,
            block_data_input_addr,
            block_data_input_size,
            signature_output_addr,
            signature_output_size,
        )

        extra_output = ""
        if response:
            output_signature_size = response
            if mboot.status_code == StatusCode.SUCCESS:
                extra_output = "Output data size/value(s) is(are):\n"
            else:
                extra_output = (
                    "Output buffer(s) is(are) smaller than the minimum requested which is(are):\n"
                )
            extra_output += (
                f"\tSignature size: {output_signature_size} ({hex(output_signature_size)})"
            )
        display_output(
            [response], mboot.status_code, ctx.obj["use_json"], ctx.obj["silent"], extra_output
        )


@trust_provisioning.command(name="oem_gen_master_share", cls=OemGenMasterShareHelp)
@click.argument("oem_share_input_addr", type=INT(), required=True)
@click.argument("oem_share_input_size", type=INT(), required=True)
@click.argument("oem_enc_share_output_addr", type=INT(), required=True)
@click.argument("oem_enc_share_output_size", type=INT(), required=True)
@click.argument("oem_enc_master_share_output_addr", type=INT(), required=True)
@click.argument("oem_enc_master_share_output_size", type=INT(), required=True)
@click.argument("oem_cust_cert_puk_output_addr", type=INT(), required=True)
@click.argument("oem_cust_cert_puk_output_size", type=INT(), required=True)
@click.pass_context
def oem_gen_master_share(
    ctx: click.Context,
    oem_share_input_addr: int,
    oem_share_input_size: int,
    oem_enc_share_output_addr: int,
    oem_enc_share_output_size: int,
    oem_enc_master_share_output_addr: int,
    oem_enc_master_share_output_size: int,
    oem_cust_cert_puk_output_addr: int,
    oem_cust_cert_puk_output_size: int,
) -> None:
    """Creates shares for initial trust provisioning keys.

    \b
    OEM_SHARE_INPUT_ADDR            - The input buffer address where the OEM Share(entropy seed) locates at
    OEM_SHARE_INPUT_SIZE             - The byte count of the OEM Share
    OEM_ENC_SHARE_OUTPUT_ADDR        - The output buffer address where ROM writes the Encrypted OEM Share to
    OEM_ENC_SHARE_OUTPUT_SIZE        - The output buffer size in byte
    OEM_ENC_MASTER_SHARE_OUTPUT_ADDR - The output buffer address where ROM writes the Encrypted OEM Master Share to
    OEM_ENC_MASTER_SHARE_OUTPUT_SIZE - The output buffer size in byte.
    OEM_CUST_CERT_PUK_OUTPUT_ADDR    - The output buffer address where ROM writes
                                       the OEM Customer Certificate Public Key to
    OEM_CUST_CERT_PUK_OUTPUT_SIZE    - The output buffer size in byte
    """
    with McuBoot(ctx.obj["interface"]) as mboot:
        response = mboot.tp_oem_gen_master_share(
            oem_share_input_addr,
            oem_share_input_size,
            oem_enc_share_output_addr,
            oem_enc_share_output_size,
            oem_enc_master_share_output_addr,
            oem_enc_master_share_output_size,
            oem_cust_cert_puk_output_addr,
            oem_cust_cert_puk_output_size,
        )
        extra_output = ""
        if response:
            oem_enc_share_size = response[0]
            oem_enc_master_share_size = response[1]
            oem_cust_cert_puk_size = response[2]
            if mboot.status_code == StatusCode.SUCCESS:
                extra_output = "Output data size/value(s) is(are):\n"
            else:
                extra_output = (
                    "Output buffer(s) is(are) smaller than the minimum requested which is(are):\n"
                )
            extra_output += (
                f"\tOEM Share size: {oem_enc_share_size} ({hex(oem_enc_share_size)})\n"
                f"\tOEM Master Share size: {oem_enc_master_share_size} ({hex(oem_enc_master_share_size)})\n"
                f"\tCust Cert Puk size: {oem_cust_cert_puk_size} ({hex(oem_cust_cert_puk_size)})"
            )
        display_output(
            response, mboot.status_code, ctx.obj["use_json"], ctx.obj["silent"], extra_output
        )


@trust_provisioning.command(name="oem_set_master_share", cls=OemSetMasterShareHelp)
@click.argument("oem_share_input_addr", type=INT(), required=True)
@click.argument("oem_share_input_size", type=INT(), required=True)
@click.argument("oem_enc_master_share_input_addr", type=INT(), required=True)
@click.argument("oem_enc_master_share_input_size", type=INT(), required=True)
@click.pass_context
def oem_set_master_share(
    ctx: click.Context,
    oem_share_input_addr: int,
    oem_share_input_size: int,
    oem_enc_master_share_input_addr: int,
    oem_enc_master_share_input_size: int,
) -> None:
    """Takes the entropy seed and the Encrypted OEM Master Share."""
    with McuBoot(ctx.obj["interface"]) as mboot:
        mboot.tp_oem_set_master_share(
            oem_share_input_addr,
            oem_share_input_size,
            oem_enc_master_share_input_addr,
            oem_enc_master_share_input_size,
        )
        display_output([], mboot.status_code, ctx.obj["use_json"], ctx.obj["silent"])


@trust_provisioning.command(name="oem_get_cust_cert_dice_puk")
@click.argument("oem_rkth_input_addr", type=INT(), required=True)
@click.argument("oem_rkth_input_size", type=INT(), required=True)
@click.argument("oem_cust_cert_dice_puk_output_addr", type=INT(), required=True)
@click.argument("oem_cust_cert_dice_puk_output_size", type=INT(), required=True)
@click.pass_context
def oem_get_cust_cert_dice_puk(
    ctx: click.Context,
    oem_rkth_input_addr: int,
    oem_rkth_input_size: int,
    oem_cust_cert_dice_puk_output_addr: int,
    oem_cust_cert_dice_puk_output_size: int,
) -> None:
    """Creates the initial trust provisioning keys.

    \b
    OEM_RKTH_INPUT_ADDR                - The input buffer address where the OEM RKTH locates at
    OEM_RKTH_INPUT_SIZE                - The byte count of the OEM RKTH
    OEM_CUST_CERT_DICE_PUK_OUTPUT_ADDR - The output buffer address where ROM writes the OEM Customer
                                         Certificate Public Key for DICE to
    OEM_CUST_CERT_DICE_PUK_OUTPUT_SIZE - The output buffer size in byte
    """
    with McuBoot(ctx.obj["interface"]) as mboot:
        response = mboot.tp_oem_get_cust_cert_dice_puk(
            oem_rkth_input_addr,
            oem_rkth_input_size,
            oem_cust_cert_dice_puk_output_addr,
            oem_cust_cert_dice_puk_output_size,
        )
    extra_output = ""
    if response:
        output_size = response
        if mboot.status_code == StatusCode.SUCCESS:
            extra_output = "Output data size/value(s) is(are):\n"
        else:
            extra_output = (
                "Output buffer(s) is(are) smaller than the minimum requested which is(are):"
            )
        extra_output += f"\tCust Cert Dice Puk size: {output_size} ({hex(output_size)})"
    display_output(
        [response], mboot.status_code, ctx.obj["use_json"], ctx.obj["silent"], extra_output
    )


@main.command()
@click.argument("life-cycle", metavar="LIFE CYCLE", type=INT(), required=True)
@click.pass_context
def update_life_cycle(ctx: click.Context, life_cycle: int) -> None:
    """Update life cycle of device.

    \b
    LIFE CYCLE    - Device life cycle to be device move to.
    """
    with McuBoot(ctx.obj["interface"]) as mboot:
        mboot.update_life_cycle(life_cycle)
        display_output([], mboot.status_code, ctx.obj["use_json"], ctx.obj["silent"])


@main.command()
@click.argument("cmd-msg-addr", metavar="COMMAND MESSAGE ADDRESS", type=INT(), required=True)
@click.argument("cmd-msg-cnt", metavar="COMMAND MESSAGE COUNT", type=INT(), required=True)
@click.argument("resp-msg-addr", metavar="RESPONSE MESSAGE ADDRESS", type=INT(), required=True)
@click.argument("resp-msg-cnt", metavar="RESPONSE MESSAGE COUNT", type=INT(), required=True)
@click.pass_context
def ele_message(
    ctx: click.Context, cmd_msg_addr: int, cmd_msg_cnt: int, resp_msg_addr: int, resp_msg_cnt: int
) -> None:
    """Send message to EdgeLock Enclave.

    This command is designed to be, as general, as is possible to work with EdgeLock Enclave.
    EdgeLock Enclave message is prepared in PC and stored in target RAM (for example by 'blhost write-memory').
    The response of ELE command is stored also in target memory on place that is defined by 'resp-msg-addr
    and could be read back (for example by 'blhost read-memory').


    Size of command message and response is in count of 32-bit words.

    \b
    COMMAND MESSAGE ADDRESS     - Address in target where is stored the command words.
    COMMAND MESSAGE COUNT       - Count of the stored command words.
    RESPONSE MESSAGE ADDRESS    - Address in target memory space where the ELE store response.
    RESPONSE MESSAGE COUNT      - Maximal count of words reserved for response.
    """
    with McuBoot(ctx.obj["interface"]) as mboot:
        mboot.ele_message(
            cmdMsgAddr=cmd_msg_addr,
            cmdMsgCnt=cmd_msg_cnt,
            respMsgAddr=resp_msg_addr,
            respMsgCnt=resp_msg_cnt,
        )
        display_output([], mboot.status_code, ctx.obj["use_json"], ctx.obj["silent"])


@trust_provisioning.command(name="prove_genuinity")
@click.argument("address", type=INT())
@click.argument("buffer_size", type=INT())
@click.pass_context
def prove_genuinity(ctx: click.Context, address: int, buffer_size: int) -> None:
    """Start the process of proving genuinity.

    \b
    ADDRESS     - Address where is the prove_genuinity request stored
    BUFFER_SIZE - Maximal size of the generated prove_genuinity response
    """
    with McuBoot(ctx.obj["interface"]) as mboot:
        tp_response_length = mboot.tp_prove_genuinity(address=address, buffer_size=buffer_size)
        display_output(
            [tp_response_length],
            mboot.status_code,
            use_json=ctx.obj["use_json"],
            extra_output=f"TP response will be {tp_response_length} bytes long."
            if tp_response_length
            else None,
        )


@trust_provisioning.command(name="isp_set_wrap_data")
@click.argument("address", type=INT())
@click.argument("control", type=INT(), required=False, default=0x1)
@click.argument("stage", type=INT(), required=False, default=0x4B)
@click.pass_context
def set_wrap_data(ctx: click.Context, address: int, control: int, stage: int) -> None:
    """Start the process of setting wrapped OEM data.

    \b
    ADDRESS - Address of the Wrapped data package in target
    CONTROL - Controls location of the Wrapped data package (1 - by address /default/, 2 - in firmware)
    STAGE   - Stage of the OEM TrustProvisioning process
    """
    with McuBoot(ctx.obj["interface"]) as mboot:
        mboot.tp_set_wrapped_data(address=address, control=control, stage=stage)
        display_output(None, mboot.status_code, use_json=ctx.obj["use_json"])


def display_output(
    response: Optional[list] = None,
    status_code: int = 0,
    use_json: bool = False,
    suppress: bool = False,
    extra_output: Optional[str] = None,
) -> None:
    """Displays response and status code.

    :param response: Response from the MBoot function
    :param status_code: MBoot status code
    :param use_json: Format the output in JSON format, defaults to False
    :param suppress: Suppress display
    :param extra_output: Extra string to print out, defaults to None
    :raises SPSDKAppError: Command is executed properly, how MBoot status code is non-zero
    """
    if suppress:
        pass
    elif use_json:
        data = {
            # get the name of a caller function and replace _ with -
            "command": inspect.stack()[1].function.replace("_", "-"),
            # this is just a visualization thing
            "response": response or [],
            "status": {
                "description": stringify_status_code(status_code),
                "value": status_code,
            },
        }
        print(json.dumps(data, indent=3))
    else:
        print(f"Response status = {stringify_status_code(status_code)}")
        if isinstance(response, list):
            filtered_response = filter(lambda x: x is not None, response)
            for i, word in enumerate(filtered_response):
                print(f"Response word {i + 1} = {word} ({word:#x})")
        if extra_output:
            print(extra_output)
    # Force exit to handover the current status code.
    # We could do that because this function is called as last from each subcommand
    if status_code:
        raise SPSDKAppError()


# For backward compatibility
decode_status_code = stringify_status_code


@catch_spsdk_error
def safe_main() -> None:
    """Calls the main function."""
    sys.exit(main())  # pylint: disable=no-value-for-parameter


if __name__ == "__main__":
    safe_main()



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2021-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""Console script for Trust provisioning config application."""
import os
import sys
from typing import List

import click

from spsdk.apps.tp_utils import (
    TPConfigConfig,
    device_help,
    get_counters,
    list_tpdevices,
    multiple_tp_dict,
    process_tp_inputs,
    tp_device_options,
)
from spsdk.apps.utils import (
    CommandsTreeGroupAliasedGetCfgTemplate,
    catch_spsdk_error,
    spsdk_apps_common_options,
    spsdk_logger,
)
from spsdk.tp import TP_DATA_FOLDER, SPSDKTpError, TpDevInterface, TrustProvisioningConfig
from spsdk.tp.utils import get_supported_devices, scan_tp_devices
from spsdk.utils.misc import load_text, write_file


@click.group(name="tpconfig", cls=CommandsTreeGroupAliasedGetCfgTemplate)
@spsdk_apps_common_options
def main(log_level: int) -> int:
    """Application for configuration of trusted device."""
    spsdk_logger.install(level=log_level)
    return 0


@main.command(name="load", no_args_is_help=True)
@tp_device_options
@click.option(
    "-t",
    "--timeout",
    type=click.IntRange(0, 600, clamp=True),
    help="The target provisioning timeout in seconds.",
)
@click.option(
    "-c",
    "--config",
    type=click.Path(exists=True, dir_okay=False),
    help="Path to configuration file (parameters on CLI take precedence).",
    required=True,
)
@click.option(
    "-s",
    "--seal",
    "seal_flag",
    is_flag=True,
    default=False,
    help="""
        Seal the smart card (advancing its lifecycle to PRODUCTION).
        WARNING: You'll not be able to use tpconfig afterwards!""",
)
def load(
    tp_device: str,
    tp_device_parameter: List[str],
    timeout: int,
    config: str,
    seal_flag: bool,
) -> None:
    """Command to load configuration to the TP device."""
    tp_config = TPConfigConfig(config, tp_device, tp_device_parameter, timeout)

    tp_interface = process_tp_inputs(
        tp_type=tp_config.tp_device,
        tp_parameters=tp_config.tp_device_parameter,
        header="device",
        scan_func=scan_tp_devices,
        print_func=click.echo,
    )
    tp_dev = tp_interface.create_interface()
    assert isinstance(tp_dev, TpDevInterface)

    tp_worker = TrustProvisioningConfig(tp_dev, click.echo)
    tp_worker.upload(tp_config.config_data, tp_config.config_dir, timeout=tp_config.timeout)
    if seal_flag:
        tp_worker.seal(timeout=tp_config.timeout)


@main.command(name="seal", no_args_is_help=True)
@tp_device_options
@click.option(
    "-t",
    "--timeout",
    type=click.IntRange(0, 600, clamp=True),
    help="The target provisioning timeout in seconds.",
)
@click.option(
    "-c",
    "--config",
    type=click.Path(exists=True, dir_okay=False),
    help="Path to configuration file (parameters on CLI take precedence).",
    required=False,
)
def seal(
    tp_device: str,
    tp_device_parameter: List[str],
    timeout: int,
    config: str,
) -> None:
    """Seal the smart card (advancing its lifecycle to PRODUCTION).

    WARNING: You'll not be able to use tpconfig afterwards!
    """
    if config:
        tp_config = TPConfigConfig(config, tp_device, tp_device_parameter, timeout)
        device = tp_config.tp_device
        params = tp_config.tp_device_parameter
        timeout_value = tp_config.timeout
    else:
        device = tp_device
        params = multiple_tp_dict(tp_device_parameter)
        timeout_value = timeout

    if not device:
        raise SPSDKTpError("TP Device's type is not specified")
    if "id" not in params or not params["id"]:
        raise SPSDKTpError("TP Device's ID is not specified")

    tp_interface = process_tp_inputs(
        tp_type=device,
        tp_parameters=params,
        header="device",
        scan_func=scan_tp_devices,
        print_func=click.echo,
    )
    tp_dev = tp_interface.create_interface()
    assert isinstance(tp_dev, TpDevInterface)

    tp_worker = TrustProvisioningConfig(tp_dev, click.echo)
    tp_worker.seal(timeout=timeout_value)


@main.command(name="get-template", no_args_is_help=True)
@click.option(
    "-f",
    "--family",
    type=click.Choice(get_supported_devices(), case_sensitive=False),
    default="lpc55s6x",
    help="Chip family to generate the TPConfig config for.",
)
@click.option(
    "-o",
    "--output",
    type=click.Path(dir_okay=False, resolve_path=True),
    required=True,
    help="The output YAML template configuration file name.",
)
# pylint: disable=unused-argument   # preparation for the future
def get_template(family: str, output: str) -> None:
    """Command to generate tphost template of configuration YML file."""
    template = load_text(os.path.join(TP_DATA_FOLDER, "tpconfig_cfg_template.yaml"))
    write_file(template, output)

    click.echo(f"The configuration template created. {output}")


main.add_command(device_help)
main.add_command(list_tpdevices)
main.add_command(get_counters)


@catch_spsdk_error
def safe_main() -> None:
    """Call the main function."""
    sys.exit(main())  # pragma: no cover  # pylint: disable=no-value-for-parameter


if __name__ == "__main__":
    safe_main()  # pragma: no cover



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2021-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""Module provides exceptions for SPSDK utilities."""
from spsdk import SPSDKError


class SPSDKRegsError(SPSDKError):
    """General Error group for utilities SPSDK registers module."""


class SPSDKRegsErrorRegisterGroupMishmash(SPSDKRegsError):
    """Register Group inconsistency problem."""


class SPSDKRegsErrorRegisterNotFound(SPSDKRegsError):
    """Register has not been found."""


class SPSDKRegsErrorBitfieldNotFound(SPSDKRegsError):
    """Bitfield has not been found."""


class SPSDKRegsErrorEnumNotFound(SPSDKRegsError):
    """Enum has not been found."""


class SPSDKTimeoutError(TimeoutError, SPSDKError):
    """SPSDK Timeout."""



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2019-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""Base for SPSDK exceptions."""
from typing import Optional

#######################################################################
# # Secure Provisioning SDK Exceptions
#######################################################################


class SPSDKError(Exception):
    """Secure Provisioning SDK Base Exception."""

    fmt = "SPSDK: {description}"

    def __init__(self, desc: Optional[str] = None) -> None:
        """Initialize the base SPSDK Exception."""
        super().__init__()
        self.description = desc

    def __str__(self) -> str:
        return self.fmt.format(description=self.description or "Unknown Error")


class SPSDKKeyError(SPSDKError, KeyError):
    """SPSDK standard key error."""


class SPSDKValueError(SPSDKError, ValueError):
    """SPSDK standard value error."""


class SPSDKTypeError(SPSDKError, TypeError):
    """SPSDK standard type error."""


class SPSDKIOError(SPSDKError, IOError):
    """SPSDK standard IO error."""


class SPSDKNotImplementedError(SPSDKError, NotImplementedError):
    """SPSDK standard not implemented error."""


class SPSDKLengthError(SPSDKError, ValueError):
    """SPSDK parsing error of any AHAB containers.

    Input/output data must be of at least container declared length bytes long.
    """


class SPSDKOverlapError(SPSDKError, ValueError):
    """Data overlap error."""


class SPSDKAlignmentError(SPSDKError, ValueError):
    """Data improperly aligned."""


class SPSDKParsingError(SPSDKError):
    """Cannot parse binary data."""


class SPSDKCorruptedException(SPSDKError):
    """Corrupted Exception."""


class SPSDKUnsupportedOperation(SPSDKError):
    """SPSDK unsupported operation error."""


class SPSDKSyntaxError(SyntaxError, SPSDKError):
    """SPSDK syntax error."""


class SPSDKFileNotFoundError(FileNotFoundError, SPSDKError):
    """SPSDK file not found error."""


class SPSDKAttributeError(SPSDKError, AttributeError):
    """SPSDK standard attribute error."""



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2021-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause
"""Module containing helper functions for elftosb."""

import struct
from numbers import Number
from typing import Callable, Dict, List, Optional, Union

from spsdk import SPSDKError
from spsdk.mboot.memories import MemId
from spsdk.sbfile.sb2.commands import (
    CmdBaseClass,
    CmdErase,
    CmdFill,
    CmdJump,
    CmdKeyStoreBackup,
    CmdKeyStoreRestore,
    CmdLoad,
    CmdMemEnable,
    CmdProg,
    CmdVersionCheck,
)
from spsdk.utils.crypto import KeyBlob
from spsdk.utils.misc import (
    align_block,
    get_bytes_cnt_of_int,
    load_binary,
    swap32,
    value_to_bytes,
    value_to_int,
)


def get_mem_id(mem_opt: Union[int, str]) -> int:
    """Get memory ID from str or int in BD file.

    :param mem_opt: memory option in BD file
    :raises SPSDKError: if memory option is not supported
    :return: int memory ID
    """
    if isinstance(mem_opt, str):
        mem_id = int(MemId.get_legacy_str(mem_opt))  # type: ignore
    elif isinstance(mem_opt, int):
        mem_id = mem_opt
    if not mem_id:
        raise SPSDKError(f"Unsupported memory option: {mem_opt}")
    return mem_id


def get_command(cmd_name: str) -> Callable[[Dict], CmdBaseClass]:
    """Returns a function based on input argument name.

    The json file generated by bd file parser uses command names (load, fill,
    etc.). These names are used to get the proper function name, which creates
    corresponding object.

    :param cmd_name: one of 'load', 'fill', 'erase', 'enable', 'reset', 'encrypt',
    'keywrap'
    :return: appropriate Command object
    """
    command_object = cmds[cmd_name]
    return command_object


def _fill_memory(cmd_args: dict) -> CmdFill:
    """Returns a CmdFill object initialized based on cmd_args.

    Fill is a type of load command used for filling a region of memory with pattern.

    Example:
    section(0) {
        // pattern fill
        load 0x55.b > 0x2000..0x3000;
        // load two bytes at an address
        load 0x1122.h > 0xf00;
    }

    :param cmd_args: dictionary holding address and pattern
    :return: CmdFill object
    """
    address = cmd_args["address"]
    pattern = cmd_args["pattern"]
    # if not isinstance(pattern, bytes):
    #     # convert to bytes
    # pattern = int.to_bytes(pattern, 4, "little")
    return CmdFill(address=address, pattern=pattern)


def _load(cmd_args: dict) -> Union[CmdLoad, CmdProg]:
    """Returns a CmdLoad object initialized based on cmd_args.

    The load statement is used to store data into the memory.
    The load command is also used to write to the flash memory.
    When loading to the flash memory, the region being loaded to must be erased before to the load operation.
    The most common form of a load statement is loading a source file by name.

    Example:
    section (0) {
        // load an entire binary file to an address
        load myBinFile > 0x70000000;
        // load an eight byte blob
        load {{ ff 2e 90 07 77 5f 1d 20 }} > 0xa0000000;
        // 4 byte load IFR statement
        load ifr 0x1234567 > 0x30;
        // Program fuse statement
        load fuse {{00 00 00 01}} > 0x01000188;
        // load to sdcard
        load sdcard {{aa bb cc dd}} > 0x08000188;
        load @288 {{aa bb cc dd}} > 0x08000188;
    }

    :param cmd_args: dictionary holding path to file or values and address
    :raises SPSDKError: If dict doesn't contain 'file' or 'values' key
    :return: CmdLoad object
    """
    prog_mem_id = 4
    address = cmd_args["address"]
    load_opt = cmd_args.get("load_opt")
    mem_id = 0
    if load_opt:
        mem_id = get_mem_id(load_opt)

    # general non-authenticated load command
    if cmd_args.get("file"):
        data = load_binary(cmd_args["file"])
        return CmdLoad(address=address, data=data, mem_id=mem_id)
    if cmd_args.get("values"):
        # if the memory ID is fuse or IFR change load command to program command
        if mem_id == prog_mem_id:
            return _prog(cmd_args, mem_id)

        values = [int(s, 16) for s in cmd_args["values"].split(",")]
        data = struct.pack(f"<{len(values)}L", *values)
        return CmdLoad(address=address, data=data, mem_id=mem_id)
    if cmd_args.get("pattern"):
        # if the memory ID is fuse or IFR change load command to program command
        # pattern in this case represents 32b int data word 1
        if mem_id == prog_mem_id:
            return _prog(cmd_args, mem_id)

    raise SPSDKError(f"Unsupported LOAD command args: {cmd_args}")


def _prog(cmd_args: dict, mem_id: int) -> CmdProg:
    """Returns a CmdProg object initialized based on cmd_args.

    :param cmd_args: dictionary holding path to file or values and address
    :param int: memory id for programming
    :raises SPSDKError: If data words are wrong
    :return: CmdProg object
    """
    address = cmd_args["address"]
    data_word1 = 0
    data_word2 = 0
    # values provided as binary blob {{aa bb cc dd}} either 4 or 8 bytes:
    if cmd_args.get("values"):
        int_value = int(cmd_args["values"], 16)
        byte_count = get_bytes_cnt_of_int(int_value)

        if byte_count <= 4:
            data_word1 = int_value
        elif byte_count <= 8:
            data_words = value_to_bytes(int_value, byte_cnt=8)
            data_word1 = value_to_int(data_words[:4])
            data_word2 = value_to_int(data_words[4:])
        else:
            raise SPSDKError("Program operation requires 4 or 8 byte segment")

        # swap byte order
        data_word1 = swap32(data_word1)
        data_word2 = swap32(data_word2)

    # values provided as integer e.g. 0x1000 represents data_word1
    elif cmd_args.get("pattern"):
        int_value = cmd_args["pattern"]
        byte_count = get_bytes_cnt_of_int(int_value)

        if byte_count <= 4:
            data_word1 = int_value
        else:
            raise SPSDKError("Data word 1 must be 4 bytes long")
    else:
        raise SPSDKError("Unsupported program command arguments")

    return CmdProg(address=address, data_word1=data_word1, data_word2=data_word2, mem_id=mem_id)


def _erase_cmd_handler(cmd_args: dict) -> CmdErase:
    """Returns a CmdErase object initialized based on cmd_args.

    The erase statement inserts a bootloader command to erase the flash memory.
    There are two forms of the erase statement. The simplest form (erase all)
    creates a command that erases the available flash memory.
    The actual effect of this command depends on the runtime settings
    of the bootloader and whether
    the bootloader resides in the flash, ROM, or RAM.

    Example:
    section (0){
        // Erase all
        erase all;
        // Erase unsecure all
        erase unsecure all;
        // erase statements specifying memory ID and range
        erase @8 all;
        erase @288 0x8001000..0x80074A4;
        erase sdcard 0x8001000..0x80074A4;
        erase mmccard 0x8001000..0x80074A4;
    }

    :param cmd_args: dictionary holding path to address, length and flags
    :return: CmdErase object
    """
    address = cmd_args["address"]
    length = cmd_args.get("length", 0)
    flags = cmd_args.get("flags", 0)

    mem_opt = cmd_args.get("mem_opt")
    mem_id = 0
    if mem_opt:
        mem_id = get_mem_id(mem_opt)

    return CmdErase(address=address, length=length, flags=flags, mem_id=mem_id)


def _enable(cmd_args: dict) -> CmdMemEnable:
    """Returns a CmdEnable object initialized based on cmd_args.

    Enable statement is used for initialization of external memories
    using a parameter block that was previously loaded to RAM.

    Example:
    section (0){
        # Load quadspi config block bin file to RAM, use it to enable QSPI.
        load myBinFile > 0x20001000;
        enable qspi 0x20001000;
    }

    :param cmd_args: dictionary holding address, size and memory type
    :return: CmdEnable object
    """
    address = cmd_args["address"]
    size = cmd_args.get("size", 4)
    mem_opt = cmd_args.get("mem_opt")
    mem_id = 0
    if mem_opt:
        mem_id = get_mem_id(mem_opt)
    return CmdMemEnable(address=address, size=size, mem_id=mem_id)


def _encrypt(cmd_args: dict) -> CmdLoad:
    """Returns a CmdLoad object initialized based on cmd_args.

    Encrypt holds an ID, which is a reference to keyblob to be used for
    encryption. So the encrypt command requires a list of keyblobs, the keyblob
    ID and load command.

    e.g.
    encrypt (0){
        load myImage > 0x0810000;
    }

    :param cmd_args: dictionary holding list of keyblobs, keyblob ID and load dict
    :raises SPSDKError: If keyblob to be used is not in the list or is invalid
    :return: CmdLoad object
    """
    keyblob_id = cmd_args["keyblob_id"]
    keyblobs = cmd_args.get("keyblobs", [])
    load_dict = cmd_args.get("load", {})

    address = load_dict["address"]

    if load_dict.get("file"):
        data = load_binary(load_dict["file"])
    if load_dict.get("values"):
        values = [int(s, 16) for s in load_dict["values"].split(",")]
        data = struct.pack(f"<{len(values)}L", *values)

    try:
        valid_keyblob = _validate_keyblob(keyblobs, keyblob_id)
    except SPSDKError as exc:
        raise SPSDKError(f"Invalid key blob {str(exc)}") from exc

    if valid_keyblob is None:
        raise SPSDKError(f"Missing keyblob {keyblob_id} for encryption.")

    start_addr = valid_keyblob["keyblob_content"][0]["start"]
    end_addr = valid_keyblob["keyblob_content"][0]["end"]
    key = bytes.fromhex(valid_keyblob["keyblob_content"][0]["key"])
    counter = bytes.fromhex(valid_keyblob["keyblob_content"][0]["counter"])
    byte_swap = valid_keyblob["keyblob_content"][0].get("byte_swap", False)

    keyblob = KeyBlob(start_addr=start_addr, end_addr=end_addr, key=key, counter_iv=counter)

    # Encrypt only if the ADE and VLD flags are set
    if bool(end_addr & keyblob.KEY_FLAG_ADE) and bool(end_addr & keyblob.KEY_FLAG_VLD):
        encoded_data = keyblob.encrypt_image(
            base_address=address, data=align_block(data, 512), byte_swap=byte_swap
        )
    else:
        encoded_data = data

    return CmdLoad(address, encoded_data)


def _keywrap(cmd_args: dict) -> CmdLoad:
    """Returns a CmdLoad object initialized based on cmd_args.

    Keywrap holds keyblob ID to be encoded by a value stored in load command and
    stored to address defined in the load command.

    Example:
    keywrap (0) {
        load {{ 00000000 }} > 0x08000000;
    }

    :param cmd_args: dictionary holding list of keyblobs, keyblob ID and load dict
    :raises SPSDKError: If keyblob to be used is not in the list or is invalid
    :return: CmdLoad object
    """
    # iterate over keyblobs
    keyblobs = cmd_args.get("keyblobs", None)
    keyblob_id = cmd_args.get("keyblob_id", None)
    load_info = cmd_args.get("load", None)

    address = load_info.get("address")
    otfad_key = load_info.get("values")

    try:
        valid_keyblob = _validate_keyblob(keyblobs, keyblob_id)
    except SPSDKError as exc:
        raise SPSDKError(f" Key blob validation failed: {str(exc)}") from exc
    if valid_keyblob is None:
        raise SPSDKError(f"Missing keyblob {keyblob_id} for given keywrap")

    start_addr = valid_keyblob["keyblob_content"][0]["start"]
    end_addr = valid_keyblob["keyblob_content"][0]["end"]
    key = bytes.fromhex(valid_keyblob["keyblob_content"][0]["key"])
    counter = bytes.fromhex(valid_keyblob["keyblob_content"][0]["counter"])

    blob = KeyBlob(start_addr=start_addr, end_addr=end_addr, key=key, counter_iv=counter)

    encoded_keyblob = blob.export(kek=otfad_key)
    print("creating wrapped keyblob")

    return CmdLoad(address=address, data=encoded_keyblob)


def _keystore_to_nv(cmd_args: dict) -> CmdKeyStoreRestore:
    """Returns a CmdKeyStoreRestore object initialized with memory type and address.

    The keystore_to_nv statement instructs the bootloader to load the backed up
    keystore values back into keystore memory region on non-volatile memory.

    Example:
    section (0) {
        keystore_to_nv @9 0x8000800;

    :param cmd_args: dictionary holding the memory type and address.
    :return: CmdKeyStoreRestore object.
    """
    mem_opt = cmd_args["mem_opt"]
    address = cmd_args["address"]
    return CmdKeyStoreRestore(address, mem_opt)


def _keystore_from_nv(cmd_args: dict) -> CmdKeyStoreBackup:
    """Returns a CmdKeyStoreRestore object initialized with memory type and address.

    The keystore_to_nv statement instructs the bootloader to load the backed up
    keystore values back into keystore memory region on non-volatile memory.

    Example:
    section (0) {
        keystore_to_nv @9 0x8000800;

    :param cmd_args: dictionary holding the memory type and address.
    :return: CmdKeyStoreRestore object.
    """
    mem_opt = cmd_args["mem_opt"]
    address = cmd_args["address"]
    return CmdKeyStoreBackup(address, mem_opt)


def _version_check(cmd_args: dict) -> CmdVersionCheck:
    """Returns a CmdVersionCheck object initialized with version check type and version.

    Validates version of secure or non-secure firmware.
    The command fails if version is < expected.

    Example:
    section (0) {
        version_check sec 0x2;
        version_check nsec 2;
    }

    :param cmd_args: dictionary holding the version type and fw version.
    :return: CmdKeyStoreRestore object.
    """
    ver_type = cmd_args["ver_type"]
    fw_version = cmd_args["fw_version"]
    return CmdVersionCheck(ver_type, fw_version)


def _validate_keyblob(keyblobs: List, keyblob_id: Number) -> Optional[Dict]:
    """Checks, whether a keyblob is valid.

    Parser returns a list of dicts which contains keyblob definitions. These
    definitions should contain a 'start', 'end', 'key' & 'counter' keys with
    appropriate values. To be able to create a keyblob, we need these for
    values. Otherwise we throw an exception that the keyblob is invalid.

    :param keyblobs: list of dicts defining keyblobs
    :param keyblob_id: id of keyblob we want to check
    :raises SPSDKError: If the keyblob definition is empty
    :raises SPSDKError: If the keyblob definition is missing one key
    :return: keyblob If exists and is valid, None otherwise
    """
    for keyblob in keyblobs:
        if keyblob_id == keyblob["keyblob_id"]:
            kb_content = keyblob["keyblob_content"]
            if len(kb_content) == 0:
                raise SPSDKError(f"Keyblob {keyblob_id} definition is empty!")

            for key in ["start", "end", "key", "counter"]:
                if key not in kb_content[0]:
                    raise SPSDKError(f"Keyblob {keyblob_id} is missing '{key}' definition!")

            return keyblob

    return None


def _jump(cmd_args: dict) -> CmdJump:
    """Returns a CmdJump object initialized with memory type and address.

    The "jump" command produces the ROM_JUMP_CMD.
    See the boot image format design document for specific details about these commands,
    such as the function prototypes they expect.

    Example:
    section (0) {
        # jump to the entrypoint
        jump mySRecFile;
        # jump to a fixed address
        jump 0xffff0000;
    }

    :param cmd_args: dictionary holding the argument and address.
    :return: CmdJump object.
    """
    argument = cmd_args.get("argument", 0)
    address = cmd_args["address"]
    spreg = cmd_args.get("spreg")

    return CmdJump(address, argument, spreg)


cmds = {
    "load": _load,
    "fill": _fill_memory,
    "erase": _erase_cmd_handler,
    "enable": _enable,
    "encrypt": _encrypt,
    "keywrap": _keywrap,
    "keystore_to_nv": _keystore_to_nv,
    "keystore_from_nv": _keystore_from_nv,
    "version_check": _version_check,
    "jump": _jump,
}



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2021-2022 NXP
#
# SPDX-License-Identifier: BSD-3-Clause
"""Module holding all supported Payload types."""

from spsdk.utils.easy_enum import Enum


class PayloadType(Enum):
    """Enumeration of all supported Payload types."""

    # fmt: off
    NXP_EPH_CHALLENGE_DATA_RND  = (0x2000, "NXP_EPH_CHALLENGE_DATA_RND", "Trust Provisioning initial challenge")
    NXP_EPH_DEVICE_KA_PUK       = (0x20A0, "NXP_EPH_DEVICE_KA_PUK", "Public key for Ephemeral key exchange")
    NXP_EPH_CARD_KA_PUK         = (0x20A2, "NXP_EPH_CARD_KA_PUK", "Public key for Ephemeral key exchange")

    TP_WRAP_DATA_IV             = (0x2001, "TP_WRAP_DATA_IV", "TP_WRAP_DATA_IV")
    TP_WRAP_DATA_TAG            = (0x2002, "TP_WRAP_DATA_TAG", "TP_WRAP_DATA_TAG")
    TP_WRAP_DATA_CIPHERTEXT     = (0xFF00, "TP_WRAP_DATA_CIPHERTEXT", "TP_WRAP_DATA_CIPHERTEXT")

    OEM_DIE_DEVATTEST_ID_PUK    = (0x20C0, "OEM_DIE_DEVATTEST_ID_PUK", "Public Key for OEM certificate creation")
    OEM_DIE_DEVATTEST_ID_CERT   = (0x20C5, "OEM_DIE_DEVATTEST_ID_CERT", "OEM DIE x509 certificate destination entry")
    OEM_DIE_DEVATTEST_ID_CERT_STD_E = (0x20C7, "OEM_DIE_DEVATTEST_ID_CERT", "OEM DIE x509 certificate standard entry")
    OEM_DIE_DEVATTEST_ID_PRK    = (0x20C6, "OEM_DIE_DEVATTEST_ID_PRK", "Private Key for singing OEM certificate")

    CUST_PROD_SB_KEK_SK         = (0x20C1, "CUST_PROD_SB_KEK_SK", "CUST_PROD_SB_KEK_SK")
    CUST_PROD_USER_KEK_SK       = (0x20C2, "CUST_PROD_SB_KEK_SK", "CUST_PROD_SB_KEK_SK")

    CUST_PROD_CFPA_DATA_SECRET  = (0x20C3, "CUST_PROD_CFPA_DATA_SECRET", "CUST_PROD_CFPA_DATA_SECRET")
    CUST_PROD_CMPA_DATA_SECRET  = (0x20C4, "CUST_PROD_CMPA_DATA_SECRET", "CUST_PROD_CMPA_DATA_SECRET")

    OEM_PROD_COUNTER            = (0x20D1, "CUST_PROD_COUNTER", "CUST_PROD_COUNTER")

    OEM_TP_LOG_HASH             = (0x20D2, "OEM_TP_LOG_HASH", "OEM_TP_LOG_HASH")
    OEM_TP_LOG_SIGN             = (0x20D3, "OEM_TP_LOG_SIGN", "OEM_TP_LOG_SIGN")
    OEM_TP_LOG_PRK              = (0x20D4, "OEM_TP_LOG_PRK", "OEM_TP_LOG_PRK")

    NXP_DIE_ID_AUTH_PUK         = (0x9966, "NXP_DIE_ID_AUTH_PUK", "NXP_DIE_ID_AUTH_PUK")
    NXP_DIE_ATTEST_AUTH_PUK     = (0x9999, "NXP_DIE_ATTEST_AUTH_PUK", "NXP_DIE_ATTEST_AUTH_PUK")
    NXP_DIE_ID_AUTH_CERT        = (0xF0F0, "NXP_DIE_ID_AUTH_CERT", "NXP_DIE_ID_AUTH_CERT")
    NXP_DIE_ECID_ID_UID         = (0x0F0F, "NXP_DIE_ECID_ID_UID", "NXP_DIE_ECID_ID_UID")
    NXP_DIE_RFC4122v4_ID_UUID   = (0x9696, "NXP_DIE_RFC4122v4_ID_UUID", "NXP_DIE_RFC4122v4_ID_UUID")



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2022-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""This module contains helper functions and modules for SPSDK apps."""

from .common_cli_options import (
    FC,
    CommandsTreeGroup,
    CommandsTreeGroupAliasedGetCfgTemplate,
    isp_interfaces,
    spsdk_apps_common_options,
)
from .utils import *



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2021-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause
"""Module used for generation SecureBinary V3.1."""
import logging
import os
from copy import deepcopy
from datetime import datetime
from struct import calcsize, pack, unpack_from
from typing import Any, Dict, List, Optional

from spsdk import SPSDKError
from spsdk.crypto.signature_provider import SignatureProvider, get_signature_provider
from spsdk.image import IMG_DATA_FOLDER, MBIMG_SCH_FILE
from spsdk.sbfile.sb31.commands import CFG_NAME_TO_CLASS, CmdSectionHeader, MainCmd
from spsdk.sbfile.sb31.functions import KeyDerivator
from spsdk.utils.crypto import CRYPTO_SCH_FILE
from spsdk.utils.crypto.abstract import BaseClass
from spsdk.utils.crypto.backend_internal import internal_backend
from spsdk.utils.crypto.cert_blocks import CertBlockV31
from spsdk.utils.database import Database
from spsdk.utils.misc import align_block, load_text, value_to_int
from spsdk.utils.schema_validator import ConfigTemplate, ValidationSchemas

logger = logging.getLogger(__name__)

SB3_SCH_FILE: str = os.path.join(IMG_DATA_FOLDER, "sch_sb3.yaml")
DATABASE_FILE = os.path.join(IMG_DATA_FOLDER, "database_sb31.yaml")


########################################################################################################################
# Secure Boot Image Class (Version 3.1)
########################################################################################################################
class SecureBinary31Header(BaseClass):
    """Header of the SecureBinary V3.1."""

    HEADER_FORMAT = "<4s2H3LQ4L16s"
    HEADER_SIZE = calcsize(HEADER_FORMAT)
    MAGIC = b"sbv3"
    FORMAT_VERSION = "3.1"
    DESCRIPTION_LENGTH = 16

    def __init__(
        self,
        firmware_version: int,
        curve_name: str,
        description: Optional[str] = None,
        timestamp: Optional[int] = None,
        is_nxp_container: bool = False,
        flags: int = 0,
    ) -> None:
        """Initialize the SecureBinary V3.1 Header.

        :param firmware_version: Firmware version (must be bigger than current CMPA record)
        :param curve_name: Name of the ECC curve used for Secure binary (secp256r1/secp384r1)
        :param description: Custom description up to 16 characters long, defaults to None
        :param timestamp: Timestamp (number of seconds since Jan 1st, 200), if None use current time
        :param is_nxp_container: NXP provisioning SB file, defaults to False
        :param flags: Flags for SB file, defaults to 0
        """
        self.flags = flags
        self.block_count = 0
        self.curve_name = curve_name
        self.block_size = self.calculate_block_size()
        self.image_type = 7 if is_nxp_container else 6
        self.firmware_version = firmware_version
        self.timestamp = timestamp or int(datetime.now().timestamp())
        self.image_total_length = self.HEADER_SIZE
        self.cert_block_offset = self.calculate_cert_block_offset()
        self.description = self._adjust_description(description)

    def _adjust_description(self, description: Optional[str] = None) -> bytes:
        """Format the description."""
        if not description:
            return bytes(self.DESCRIPTION_LENGTH)
        desc = bytes(description, encoding="ascii")
        desc = desc[: self.DESCRIPTION_LENGTH]
        desc += bytes(self.DESCRIPTION_LENGTH - len(desc))
        return desc

    def calculate_cert_block_offset(self) -> int:
        """Calculate the offset to the Certification block."""
        fixed_offset = 1 * 8 + 9 * 4 + 16
        if self.curve_name == "secp256r1":
            return fixed_offset + 32
        if self.curve_name == "secp384r1":
            return fixed_offset + 48
        raise SPSDKError(f"Invalid curve name: {self.curve_name}")

    def calculate_block_size(self) -> int:
        """Calculate the the data block size."""
        fixed_block_size = 4 + 256
        if self.curve_name == "secp256r1":
            return fixed_block_size + 32
        if self.curve_name == "secp384r1":
            return fixed_block_size + 48
        raise SPSDKError(f"Invalid curve name: {self.curve_name}")

    def info(self) -> str:
        """Get info of SB v31 as a string."""
        info = str()
        info += f" Magic:                       {self.MAGIC.decode('ascii')}\n"
        info += f" Version:                     {self.FORMAT_VERSION}\n"
        info += f" Flags:                       0x{self.flags:04X}\n"
        info += f" Block count:                 {self.block_count}\n"
        info += f" Block size:                  {self.block_size}\n"
        info += f" Firmware version:            {self.firmware_version}\n"
        info += f" Image type:                  {self.image_type}\n"
        info += f" Timestamp:                   {self.timestamp}\n"
        info += f" Total length of Block#0:     {self.image_total_length}\n"
        info += f" Certificate block offset:    {self.cert_block_offset}\n"
        info += f" Description:                 {self.description.decode('ascii')}\n"
        return info

    def update(self, commands: "SecureBinary31Commands", cert_block: CertBlockV31) -> None:
        """Updates the volatile fields in header by real commands and certification block data.

        :param commands: SB3.1 Commands block
        :param cert_block: SB3.1 Certification block.
        """
        self.block_count = commands.block_count
        self.image_total_length += len(commands.final_hash) + cert_block.expected_size
        self.image_total_length += 2 * len(commands.final_hash)

    def export(self) -> bytes:
        """Serialize the SB file to bytes."""
        major_format_version, minor_format_version = [
            int(v) for v in self.FORMAT_VERSION.split(".")
        ]
        return pack(
            self.HEADER_FORMAT,
            self.MAGIC,
            minor_format_version,
            major_format_version,
            self.flags,
            self.block_count,
            self.block_size,
            self.timestamp,
            self.firmware_version,
            self.image_total_length,
            self.image_type,
            self.cert_block_offset,
            self.description,
        )

    @classmethod
    def parse(cls, data: bytes, offset: int = 0) -> "SecureBinary31Header":
        """Parse binary data into SecureBinary31Header.

        :raises SPSDKError: Unable to parse SB31 Header.
        """
        (
            magic,
            minor_version,
            major_version,
            flags,
            block_count,
            block_size,
            timestamp,
            firmware_version,
            image_total_length,
            image_type,
            cert_block_offset,
            description,
        ) = unpack_from(cls.HEADER_FORMAT, data, offset=offset)
        if magic != cls.MAGIC:
            raise SPSDKError("Magic doesn't match")
        if major_version != 3 and minor_version != 1:
            raise SPSDKError(f"Unable to parse SB version {major_version}.{minor_version}")
        if block_size not in [292, 308]:
            raise SPSDKError(f"Unable to determine curve name from block size: {block_size}")

        curve_name = "secp256r1" if block_size == 292 else "secp384r1"
        obj = SecureBinary31Header(
            firmware_version=firmware_version,
            curve_name=curve_name,
            description=description.decode("utf-8"),
            timestamp=timestamp,
            is_nxp_container=image_type == 7,
            flags=flags,
        )
        obj.block_count = block_count
        obj.block_size = block_size
        obj.cert_block_offset = cert_block_offset
        obj.image_total_length = image_total_length
        return obj

    def validate(self) -> None:
        """Validate the settings of class members.

        :raises SPSDKError: Invalid configuration of SB3.1 header blob class members.
        """
        if self.flags is None:
            raise SPSDKError("Invalid SB3.1 header flags.")
        if self.block_count is None or self.block_count < 0:
            raise SPSDKError("Invalid SB3.1 header block count.")
        if self.curve_name is None or self.curve_name not in ["secp256r1", "secp384r1"]:
            raise SPSDKError("Invalid SB3.1 header curve name.")
        if self.block_size is None or self.block_size != self.calculate_block_size():
            raise SPSDKError("Invalid SB3.1 header block size.")
        if self.image_type is None or self.image_type not in [6, 7]:
            raise SPSDKError("Invalid SB3.1 header image type.")
        if self.firmware_version is None:
            raise SPSDKError("Invalid SB3.1 header firmware version.")
        if self.timestamp is None:
            raise SPSDKError("Invalid SB3.1 header timestamp.")
        if self.image_total_length is None or self.image_total_length < self.HEADER_SIZE:
            raise SPSDKError("Invalid SB3.1 header image total length.")
        if self.cert_block_offset is None:
            raise SPSDKError("Invalid SB3.1 header certification block offset.")
        if self.description is None or len(self.description) != 16:
            raise SPSDKError("Invalid SB3.1 header image description.")


class SecureBinary31Commands(BaseClass):
    """Blob containing SB3.1 commands."""

    DATA_CHUNK_LENGTH = 256

    def __init__(
        self,
        family: str,
        curve_name: str,
        is_encrypted: bool = True,
        pck: Optional[bytes] = None,
        timestamp: Optional[int] = None,
        kdk_access_rights: Optional[int] = None,
    ) -> None:
        """Initialize container for SB3.1 commands.

        :param curve_name: Name of the ECC curve used for Secure binary (secp256r1/secp384r1)
        :param is_encrypted: Indicate whether commands should be encrypted or not, defaults to True
        :param pck: Part Common Key (needed if `is_encrypted` is True), defaults to None
        :param timestamp: Timestamp used for encryption (needed if `is_encrypted` is True), defaults to None
        :param kdk_access_rights: Key Derivation Key access rights (needed if `is_encrypted` is True), defaults to None
        :raises SPSDKError: Key derivation arguments are not provided if `is_encrypted` is True
        """
        super().__init__()
        self.family = family
        self.curve_name = curve_name
        self.hash_type = self._get_hash_type(curve_name)
        self.is_encrypted = is_encrypted
        self.block_count = 0
        self.final_hash = bytes(self._get_hash_length(curve_name))
        self.commands: List[MainCmd] = []
        self.key_derivator = None
        if is_encrypted:
            if pck is None or timestamp is None or kdk_access_rights is None:
                raise SPSDKError("PCK, timestamp or kdk_access_rights are not defined.")
            self.key_derivator = KeyDerivator(
                pck=pck,
                timestamp=timestamp,
                key_length=self._get_key_length(curve_name),
                kdk_access_rights=kdk_access_rights,
            )

    @staticmethod
    def _get_hash_length(curve_name: str) -> int:
        return {"secp256r1": 32, "secp384r1": 48}[curve_name]

    @staticmethod
    def _get_key_length(curve_name: str) -> int:
        return {"secp256r1": 128, "secp384r1": 256}[curve_name]

    @staticmethod
    def _get_hash_type(curve_name: str) -> str:
        return {"secp256r1": "sha256", "secp384r1": "sha384"}[curve_name]

    def add_command(self, command: MainCmd) -> None:
        """Add SB3.1 command."""
        self.commands.append(command)

    def insert_command(self, index: int, command: MainCmd) -> None:
        """Insert SB3.1 command."""
        if index == -1:
            self.commands.append(command)
        else:
            self.commands.insert(index, command)

    def set_commands(self, commands: List[MainCmd]) -> None:
        """Set all SB3.1 commands at once."""
        self.commands = commands.copy()

    def load_from_config(
        self, config: List[Dict[str, Any]], search_paths: Optional[List[str]] = None
    ) -> None:
        """Load configuration from dictionary.

        :param config: Dictionary with configuration fields.
        :param search_paths: List of paths where to search for the file, defaults to None
        """
        for cfg_cmd in config:
            cfg_cmd_key = list(cfg_cmd.keys())[0]
            cfg_cmd_value = cfg_cmd[cfg_cmd_key]
            cfg_cmd_value["family"] = self.family
            self.add_command(
                CFG_NAME_TO_CLASS[cfg_cmd_key].load_from_config(
                    cfg_cmd_value, search_paths=search_paths
                )
            )

    def get_cmd_blocks_to_export(self) -> List[bytes]:
        """Export commands as bytes."""
        commands_bytes = b"".join([command.export() for command in self.commands])
        section_header = CmdSectionHeader(length=len(commands_bytes))
        total = section_header.export() + commands_bytes

        data_blocks = [
            total[i : i + self.DATA_CHUNK_LENGTH]
            for i in range(0, len(total), self.DATA_CHUNK_LENGTH)
        ]
        data_blocks[-1] = align_block(data_blocks[-1], alignment=self.DATA_CHUNK_LENGTH)

        return data_blocks

    def process_cmd_blocks_to_export(self, data_blocks: List[bytes]) -> bytes:
        """Process given data blocks for export."""
        self.block_count = len(data_blocks)

        processed_blocks = [
            self._process_block(block_number, block_data)
            for block_number, block_data in reversed(list(enumerate(data_blocks, start=1)))
        ]
        final_data = b"".join(reversed(processed_blocks))
        return final_data

    def export(self) -> bytes:
        """Export commands as bytes."""
        data_blocks = self.get_cmd_blocks_to_export()
        return self.process_cmd_blocks_to_export(data_blocks)

    def _process_block(self, block_number: int, block_data: bytes) -> bytes:
        """Process single block."""
        if self.is_encrypted:
            if not self.key_derivator:
                raise SPSDKError("No key derivator")
            block_key = self.key_derivator.get_block_key(block_number)
            encrypted_block = internal_backend.aes_cbc_encrypt(block_key, block_data)
        else:
            encrypted_block = block_data

        full_block = pack(
            f"<L{len(self.final_hash)}s{len(encrypted_block)}s",
            block_number,
            self.final_hash,
            encrypted_block,
        )
        block_hash = internal_backend.hash(full_block, self.hash_type)
        self.final_hash = block_hash
        return full_block

    def info(self) -> str:
        """Get string information for commands in the container."""
        info = str()
        info += "COMMANDS:\n"
        info += f"Number of commands: {len(self.commands)}\n"
        for command in self.commands:
            info += f"  {command.info()}\n"
        return info

    @classmethod
    def parse(cls, data: bytes, offset: int = 0) -> "SecureBinary31Commands":
        """Parse binary data into SecureBinary31Commands.

        :raises NotImplementedError: Not yet implemented
        """
        raise NotImplementedError("Not yet implemented.")

    def validate(self) -> None:
        """Validate the settings of class members.

        :raises SPSDKError: Invalid configuration of SB3.1 commands blob class members.
        """
        if self.is_encrypted and not self.key_derivator:
            raise SPSDKError("Invalid key derivator")


class SecureBinary31(BaseClass):
    """Secure Binary SB3.1 class."""

    def __init__(
        self,
        family: str,
        curve_name: str,
        cert_block: CertBlockV31,
        firmware_version: int,
        signature_provider: SignatureProvider,
        pck: Optional[bytes] = None,
        kdk_access_rights: Optional[int] = None,
        description: Optional[str] = None,
        is_nxp_container: bool = False,
        flags: int = 0,
        timestamp: Optional[int] = None,
        is_encrypted: bool = True,
    ) -> None:
        """Constructor for Secure Binary v3.1 data container.

        :param curve_name: Name of the ECC curve used for Secure binary (secp256r1/secp384r1).
        :param cert_block: Certification block.
        :param firmware_version: Firmware version (must be bigger than current CMPA record).
        :param signature_provider: Signature provider for final sign of SB3.1 image.
        :param pck: Part Common Key (needed if `is_encrypted` is True), defaults to None
        :param kdk_access_rights: Key Derivation Key access rights (needed if `is_encrypted` is True), defaults to None
        :param description: Custom description up to 16 characters long, defaults to None
        :param is_nxp_container: NXP provisioning SB file, defaults to False
        :param flags: Flags for SB file, defaults to 0
        :param timestamp: Timestamp used for encryption (needed if `is_encrypted` is True), defaults to None
        :param is_encrypted: Indicate whether commands should be encrypted or not, defaults to True
        """
        # in our case, timestamp is the number of seconds since "Jan 1, 2000"
        self.family = family
        self.timestamp = timestamp or int((datetime.now() - datetime(2000, 1, 1)).total_seconds())
        self.pck = pck
        self.cert_block: CertBlockV31 = cert_block
        self.curve_name = curve_name
        self.is_encrypted = is_encrypted
        self.kdk_access_rights = kdk_access_rights
        self.firmware_version = firmware_version
        self.description = description
        self.is_nxp_container = is_nxp_container
        self.flags = flags
        self.signature_provider = signature_provider

        self.sb_header = SecureBinary31Header(
            firmware_version=self.firmware_version,
            curve_name=self.curve_name,
            description=self.description,
            timestamp=self.timestamp,
            is_nxp_container=self.is_nxp_container,
            flags=self.flags,
        )
        self.sb_commands = SecureBinary31Commands(
            family=self.family,
            curve_name=curve_name,
            is_encrypted=self.is_encrypted,
            pck=pck,
            timestamp=self.timestamp,
            kdk_access_rights=self.kdk_access_rights,
        )
        if self.pck:
            logger.info(f"SB3KDK: {self.pck.hex()}")

    @staticmethod
    def _get_prv_key_length(curve_name: str) -> int:
        """Get size of key for curve in bits."""
        return {"secp256r1": 256, "secp384r1": 384}[curve_name]

    @classmethod
    def get_validation_schemas_family(cls) -> List[Dict[str, Any]]:
        """Create the validation schema just for supported families.

        :return: List of validation schemas for SB31 supported families.
        """
        sch_cfg = ValidationSchemas.get_schema_file(SB3_SCH_FILE)
        sch_cfg["sb3_family"]["properties"]["family"]["enum"] = cls.get_supported_families()
        return [sch_cfg["sb3_family"]]

    @classmethod
    def get_commands_validation_schemas(cls, family: str) -> List[Dict[str, Any]]:
        """Create the list of validation schemas.

        :param family: Family description.
        :return: List of validation schemas.
        """
        sb3_sch_cfg = ValidationSchemas().get_schema_file(SB3_SCH_FILE)

        schemas: List[Dict[str, Any]] = [deepcopy(sb3_sch_cfg["sb3_commands"])]

        # remove unused command for current family
        supported_commands = Database(DATABASE_FILE).get_device_value("supported_commands", family)
        list_of_commands: List[Dict] = schemas[0]["properties"]["commands"]["items"]["oneOf"]
        for command in list_of_commands:
            if list(command["properties"].keys())[0] not in supported_commands:
                list_of_commands.remove(command)

        return schemas

    @classmethod
    def get_validation_schemas(
        cls, family: str, include_test_configuration: bool = False
    ) -> List[Dict[str, Any]]:
        """Create the list of validation schemas.

        :param family: Family description.
        :param include_test_configuration: Add also testing configuration schemas.
        :return: List of validation schemas.
        """
        mbi_sch_cfg = ValidationSchemas().get_schema_file(MBIMG_SCH_FILE)
        crypto_sch_cfg = ValidationSchemas().get_schema_file(CRYPTO_SCH_FILE)
        sb3_sch_cfg = ValidationSchemas().get_schema_file(SB3_SCH_FILE)

        schemas: List[Dict[str, Any]] = []
        schemas.extend(
            [
                mbi_sch_cfg[x]
                for x in [
                    "firmware_version",
                    "signing_prv_key_lpc55s3x",
                    "elliptic_curves",
                ]
            ]
        )
        schemas.extend([crypto_sch_cfg[x] for x in ["certificate_v31", "certificate_root_keys"]])
        schemas.extend([sb3_sch_cfg[x] for x in ["sb3_family", "sb3", "sb3_description"]])
        schemas.extend(cls.get_commands_validation_schemas(family))
        if include_test_configuration:
            schemas.append(sb3_sch_cfg["sb3_test"])

        # find family
        for schema in schemas:
            if "properties" in schema and "family" in schema["properties"]:
                schema["properties"]["family"]["enum"] = cls.get_supported_families()
                break
        return schemas

    @classmethod
    def load_from_config(
        cls, config: Dict[str, Any], search_paths: Optional[List[str]] = None
    ) -> "SecureBinary31":
        """Creates an instance of SecureBinary31 from configuration.

        :param config: Input standard configuration.
        :param search_paths: List of paths where to search for the file, defaults to None
        :return: Instance of Secure Binary V3.1 class
        """
        family = config["family"]
        container_keyblob_enc_key_path = config.get("containerKeyBlobEncryptionKey")
        is_nxp_container = config.get("isNxpContainer", False)
        description = config.get("description")
        kdk_access_rights = value_to_int(config.get("kdkAccessRights", 0))
        container_configuration_word = value_to_int(config.get("containerConfigurationWord", 0))
        firmware_version = value_to_int(config.get("firmwareVersion", 1))

        commands = config["commands"]
        is_encrypted = config.get("isEncrypted", True)
        timestamp = config.get("timestamp")
        if timestamp:  # re-format it
            timestamp = value_to_int(timestamp)

        cert_block = CertBlockV31.from_config(config, search_paths=search_paths)

        # if use_isk is set, we use for signing the ISK certificate instead of root
        signing_key_path = (
            config.get("signingCertificatePrivateKeyFile")
            if cert_block.isk_certificate
            else config.get("mainRootCertPrivateKeyFile")
        )
        curve_name = (
            config.get("iskCertificateEllipticCurve")
            if cert_block.isk_certificate
            else config.get("rootCertificateEllipticCurve")
        )
        assert curve_name and isinstance(curve_name, str)
        assert signing_key_path
        signature_provider = get_signature_provider(
            sp_cfg=config.get("signProvider"),
            local_file_key=signing_key_path,
            search_paths=search_paths,
            mode="deterministic-rfc6979",
        )
        assert signature_provider

        pck = None
        if is_encrypted:
            assert container_keyblob_enc_key_path
            pck = bytes.fromhex(
                load_text(container_keyblob_enc_key_path, search_paths=search_paths)
            )

        # Create SB3 object
        sb3 = SecureBinary31(
            family=family,
            pck=pck,
            cert_block=cert_block,
            curve_name=curve_name,
            kdk_access_rights=kdk_access_rights,
            firmware_version=firmware_version,
            description=description,
            is_nxp_container=is_nxp_container,
            flags=container_configuration_word,
            signature_provider=signature_provider,
            timestamp=timestamp,
            is_encrypted=is_encrypted,
        )

        # Add commands into the SB3 object
        sb3.sb_commands.load_from_config(commands, search_paths=search_paths)

        return sb3

    def validate(self) -> None:
        """Validate the settings of class members.

        :raises SPSDKError: Invalid configuration of SB3.1 class members.
        """
        if self.signature_provider is None or not isinstance(
            self.signature_provider, SignatureProvider
        ):
            raise SPSDKError(f"SB3.1 signature provider is invalid: {self.signature_provider}")

        try:
            prv_key_length = self._get_prv_key_length(self.curve_name)
        except KeyError as exc:
            raise SPSDKError(f"Invalid SB3 curve name: ({self.curve_name})") from exc

        if (self.signature_provider.signature_length / 2) * 8 != prv_key_length:
            raise SPSDKError(
                f"Invalid length of SB3.1 signing key({self.signature_provider.signature_length} != {prv_key_length})"
                f" for used curve: {self.curve_name}!"
            )
        self.cert_block.validate()
        self.sb_header.validate()
        self.sb_commands.validate()

    def export(self, cert_block: Optional[bytes] = None) -> bytes:
        """Generate binary output of SB3.1 file.

        :return: Content of SB3.1 file in bytes.
        """
        self.validate()

        if cert_block:
            cert_block_data = cert_block
        else:
            cert_block_data = self.cert_block.export()
        sb3_commands_data = self.sb_commands.export()

        final_data = bytes()
        # HEADER OF SB 3.1 FILE
        self.sb_header.update(self.sb_commands, self.cert_block)
        final_data += self.sb_header.export()

        # HASH OF PREVIOUS BLOCK
        final_data += self.sb_commands.final_hash
        final_data += cert_block_data

        # SIGNATURE
        final_data += self.signature_provider.sign(final_data)

        # COMMANDS BLOBS DATA
        final_data += sb3_commands_data

        return final_data

    def info(self) -> str:
        """Create string information about SB3.1 loaded file.

        :return: Text information about SB3.1.
        """
        self.validate()
        ret = ""

        ret += "SB3.1 header:\n"
        ret += self.sb_header.info()

        ret += "SB3.1 commands blob :\n"
        ret += self.sb_commands.info()

        return ret

    @staticmethod
    def get_supported_families() -> List[str]:
        """Return list of supported families.

        :return: List of supported families.
        """
        return Database(DATABASE_FILE).devices.device_names

    @classmethod
    def generate_config_template(cls, family: str) -> Dict[str, str]:
        """Generate configuration for selected family.

        :param family: Family description.
        :return: Dictionary of individual templates (key is name of template, value is template itself).
        """
        ret: Dict[str, str] = {}

        if family in cls.get_supported_families():
            schemas = cls.get_validation_schemas(family)
            schemas.append(ValidationSchemas.get_schema_file(SB3_SCH_FILE)["sb3_output"])
            override = {}
            override["family"] = family

            yaml_data = ConfigTemplate(
                f"Secure Binary v3.1 Configuration template for {family}.",
                schemas,
                override,
            ).export_to_yaml()

            ret[f"{family}_sb31"] = yaml_data

        return ret

    @classmethod
    def parse(cls, data: bytes, offset: int = 0) -> "SecureBinary31":
        """Deserialize object from bytes array.

        :raises NotImplementedError: Not yet implemented
        """
        raise NotImplementedError("Not yet implemented.")



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2022-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""This module contains Bootable image related code."""
import os

from spsdk import SPSDK_DATA_FOLDER

BIMG_DATA_FOLDER: str = os.path.join(SPSDK_DATA_FOLDER, "image", "bootable_image")
BIMG_SCH_FILE: str = os.path.join(BIMG_DATA_FOLDER, "sch_bimg.yaml")
BIMG_DATABASE_FILE: str = os.path.join(BIMG_DATA_FOLDER, "database.yaml")



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2019-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""Internal implementation for security backend."""
import importlib
import math
from struct import pack, unpack_from
from typing import Any, Optional, Union

# Used security modules
from Crypto import Hash, Random
from Crypto.Cipher import AES
from Crypto.Hash import CMAC, HMAC
from Crypto.PublicKey import ECC, RSA
from Crypto.Signature import DSS, pkcs1_15

from spsdk import SPSDKError
from spsdk.crypto import PrivateKey

# Abstract Class Interface
from .abstract import BackendClass


########################################################################################################################
# SPSDK Backend
########################################################################################################################
class Backend(BackendClass):
    """Internal implementation for security backend."""

    @property
    def name(self) -> str:
        """Name of the backend."""
        return "SPSDK"

    @property
    def version(self) -> str:
        """Version of the backend."""
        return "0.1"

    def random_bytes(self, length: int) -> bytes:
        """Return a random byte string with specified length.

        :param length: The length in bytes
        :return: Random bytes
        """
        return Random.get_random_bytes(length)

    @staticmethod
    def _get_algorithm(name: str, data: bytes) -> Any:
        """For specified name return Hash algorithm instance.

        :param name: Name of the algorithm (class name), case insensitive
        :param data: parameter for the constructor of the algorithm class
        :return: instance of algorithm class
        :raises SPSDKError: If the algorithm is not found
        """
        # algo_cls = getattr(Hash, name.upper(), None)  # hack: get class object by name
        try:
            algo_cls = importlib.import_module(f"Crypto.Hash.{name.upper()}")
        except ModuleNotFoundError as exc:
            raise SPSDKError(f"No module named 'Crypto.Hash.{name.upper()}") from exc
        if algo_cls is None:
            raise SPSDKError(f"Unsupported algorithm: Hash.{name}".format(name=name.upper()))
        return algo_cls.new(data)  # type: ignore  # pylint: disable=not-callable

    def cmac(self, data: bytes, key: bytes) -> bytes:  # pylint: disable=no-self-use
        """Generate Cipher-based Message Authentication Code via AES.

        :param data: Data to digest
        :param key: AES Key for CMAC computation
        :return: CMAC bytes
        """
        cipher = CMAC.new(key=key, ciphermod=AES)
        cipher.update(data)
        return cipher.digest()

    def hash(self, data: bytes, algorithm: str = "sha256") -> bytes:
        """Return a HASH from input data with specified algorithm.

        :param data: Input data in bytes
        :param algorithm: Algorithm type for HASH function
        :return: Hash-ed bytes
        :raises SPSDKError: If the algorithm is not found
        """
        return self._get_algorithm(algorithm, data).digest()

    def hmac(self, key: bytes, data: bytes, algorithm: str = "sha256") -> bytes:
        """Return a HMAC from data with specified key and algorithm.

        :param key: The key in bytes format
        :param data: Input data in bytes format
        :param algorithm: Algorithm type for HASH function (sha256, sha384, sha512, ...)
        :return: HMAC bytes
        :raises SPSDKError: If the algorithm is not found
        """
        cls = getattr(Hash, algorithm.upper(), None)
        if cls is None:
            raise SPSDKError("The algorithm is not found")
        hmac_obj = HMAC.new(key, data, cls)
        return hmac_obj.digest()

    # pylint: disable=invalid-name
    def aes_key_wrap(self, kek: bytes, key_to_wrap: bytes) -> bytes:
        """Wraps a key using a key-encrypting key (KEK).

        :param kek: The key-encrypting key
        :param key_to_wrap: Plain data
        :return: Wrapped key
        :raises SPSDKError: Invalid length of kek or key_to_wrap
        """
        if len(kek) not in (16, 24, 32):
            raise SPSDKError("The wrapping key must be a valid AES key length")
        if len(key_to_wrap) < 16:
            raise SPSDKError("The key to wrap must be at least 16 bytes")
        if len(key_to_wrap) % 8 != 0:
            raise SPSDKError("The key to wrap must be a multiple of 8 bytes")
        iv = 0xA6A6A6A6A6A6A6A6
        n = len(key_to_wrap) // 8
        r = [b""] + [key_to_wrap[i * 8 : i * 8 + 8] for i in range(0, n)]
        a = iv
        aes = AES.new(kek, AES.MODE_ECB)
        for j in range(6):
            for i in range(1, n + 1):
                b = aes.encrypt(pack(">Q", a) + r[i])
                a = unpack_from(">Q", b[:8])[0] ^ (n * j + i)
                r[i] = b[8:]
        return pack(">Q", a) + b"".join(r[1:])

    # pylint: disable=invalid-name
    def aes_key_unwrap(self, kek: bytes, wrapped_key: bytes) -> bytes:
        """Unwraps a key using a key-encrypting key (KEK).

        :param kek: The key-encrypting key
        :param wrapped_key: Encrypted data
        :return: Un-wrapped key
        :raises SPSDKError: Invalid length of kek or key_to_wrap
        """
        if len(kek) not in (16, 24, 32):
            raise SPSDKError("The wrapping key must be a valid AES key length")
        if len(wrapped_key) < 24:
            raise SPSDKError("Must be at least 24 bytes")
        if len(wrapped_key) % 8 != 0:
            raise SPSDKError("The wrapped key must be a multiple of 8 bytes")
        # default iv
        iv = 0xA6A6A6A6A6A6A6A6
        n = len(wrapped_key) // 8 - 1
        # NOTE: R[0] is never accessed, left in for consistency with RFC indices
        r = [b""] + [wrapped_key[i * 8 : i * 8 + 8] for i in range(1, n + 1)]
        a = unpack_from(">Q", wrapped_key[:8])[0]
        aes = AES.new(kek, AES.MODE_ECB)
        for j in range(5, -1, -1):  # counting down
            for i in range(n, 0, -1):  # (n, n-1, ..., 1)
                b = aes.decrypt(pack(">Q", a ^ (n * j + i)) + r[i])
                a = unpack_from(">Q", b[:8])[0]
                r[i] = b[8:]
        if a != iv:
            raise SPSDKError(f"Integrity Check Failed: {a:016X} (expected {iv:016X})")
        return b"".join(r[1:])

    def aes_cbc_encrypt(
        self, key: bytes, plain_data: bytes, iv_data: Optional[bytes] = None
    ) -> bytes:
        """Encrypt plain data with AES in CBC mode.

        :param key: Key for encryption
        :param plain_data: Data to encrypt
        :param iv_data: Initial vector for encryption, defaults to None
        :return: Encrypted data
        :raises SPSDKError: Incorrect key or initialization vector size
        """
        if len(key) not in AES.key_size:
            raise SPSDKError(
                f"The key must be a valid AES key length: {', '.join([str(k) for k in AES.key_size])}"
            )
        init_vector = iv_data or bytes(AES.block_size)
        if len(init_vector) != AES.block_size:
            raise SPSDKError(f"The initial vector length must be {AES.block_size}")
        cipher = AES.new(key, mode=AES.MODE_CBC, iv=init_vector)
        return cipher.encrypt(plain_data)

    def aes_cbc_decrypt(
        self, key: bytes, encrypted_data: bytes, iv_data: Optional[bytes] = None
    ) -> bytes:
        """Decrypt encrypted data with AES in CBC mode.

        :param key: The key for data decryption
        :param encrypted_data: Input data
        :param iv_data: Initialization vector data
        :raises SPSDKError: Invalid Key or IV
        :return: Decrypted image
        """
        if len(key) not in AES.key_size:
            raise SPSDKError(
                f"The key must be a valid AES key length: {', '.join([str(k) for k in AES.key_size])}"
            )
        init_vector = iv_data or bytes(AES.block_size)
        if len(init_vector) != AES.block_size:
            raise SPSDKError(f"The initial vector length must be {AES.block_size}")
        cipher = AES.new(key, mode=AES.MODE_CBC, iv=init_vector)
        return cipher.decrypt(encrypted_data)

    def aes_ctr_encrypt(self, key: bytes, plain_data: bytes, nonce: bytes) -> bytes:
        """Encrypt plain data with AES in CTR mode.

        :param key: The key for data encryption
        :param plain_data: Input data
        :param nonce: Nonce data with counter value
        :return: Encrypted data
        :raises SPSDKError: Invalid length of key
        :raises SPSDKError: Invalid length of nonce
        :raises SPSDKError: Invalid length of plain text
        """
        if len(key) not in (16, 24, 32):
            raise SPSDKError("The key must be a valid AES key length")
        if len(nonce) != 16:
            raise SPSDKError("The nonce length is not valid")
        if len(plain_data) > len(nonce):
            raise SPSDKError("The length of plain text is large than the length of nonce")
        aes = AES.new(key, AES.MODE_ECB)
        ctr = aes.encrypt(nonce)
        return bytes([p ^ c for p, c in zip(plain_data, ctr)])

    def aes_ctr_decrypt(self, key: bytes, encrypted_data: bytes, nonce: bytes) -> bytes:
        """Decrypt encrypted data with AES in CTR mode.

        :param key: The key for data decryption
        :param encrypted_data: Input data
        :param nonce: Nonce data with counter value
        :return: Decrypted data
        """
        return self.aes_ctr_encrypt(key, encrypted_data, nonce)

    def aes_xts_encrypt(self, key: bytes, plain_data: bytes, tweak: bytes) -> bytes:
        """Encrypt plain data with AES in XTS mode.

        :param key: The key for data encryption
        :param plain_data: Input data
        :param tweak: The tweak is a 16 byte value
        :return: Encrypted data
        :raises NotImplementedError: This backend doesn't support AES-XTS
        """
        raise NotImplementedError("This backend doesn't support AES-XTS")

    def aes_xts_decrypt(self, key: bytes, encrypted_data: bytes, tweak: bytes) -> bytes:
        """Decrypt encrypted data with AES in XTS mode.

        :param key: The key for data decryption
        :param encrypted_data: Input data
        :param tweak: The tweak is a 16 byte value
        :return: Decrypted data
        :raises NotImplementedError: This backend doesn't support AES-XTS
        """
        raise NotImplementedError("This backend doesn't support AES-XTS")

    def rsa_sign(
        self,
        private_key: Union[RSA.RsaKey, bytes],
        data: bytes,
        algorithm: str = "sha256",
    ) -> bytes:
        """Sign input data.

        :param private_key: The private key: either RSA.RsaKey or decrypted binary data in PEM format
        :param data: Input data
        :param algorithm: Used algorithm
        :return: Singed data
        :raises SPSDKError: If the algorithm is not found
        """
        if isinstance(private_key, bytes):
            private_key = RSA.import_key(private_key)
        assert isinstance(private_key, RSA.RsaKey)
        h = self._get_algorithm(algorithm, data)
        return pkcs1_15.new(private_key).sign(h)

    def rsa_verify(
        self,
        pub_key_mod: int,
        pub_key_exp: int,
        signature: bytes,
        data: bytes,
        algorithm: str = "sha256",
    ) -> bool:
        """Verify input data.

        :param pub_key_mod: The public key modulus
        :param pub_key_exp: The public key exponent
        :param signature: The signature of input data
        :param data: Input data
        :param algorithm: Used algorithm
        :return: True if signature is valid, False otherwise
        :raises SPSDKError: If the algorithm is not found
        """
        public_key = self.rsa_public_key(pub_key_mod, pub_key_exp)
        assert isinstance(public_key, RSA.RsaKey)
        h = self._get_algorithm(algorithm, data)
        try:
            pkcs1_15.new(public_key).verify(h, signature)
        except (ValueError, TypeError):
            return False

        return True

    def rsa_public_key(self, modulus: int, exponent: int) -> RSA.RsaKey:
        """Create RSA public key object from modulus and exponent.

        :param modulus: The RSA public key modulus
        :param exponent: The RSA public key exponent
        :return: RSA Key instance
        """
        return RSA.construct((modulus, exponent))

    def ecc_sign(
        self, private_key: Union[ECC.EccKey, bytes], data: bytes, algorithm: Optional[str] = None
    ) -> bytes:
        """Sign data using (EC)DSA.

        :param private_key: ECC private key, either as EccKey or bytes
        :param data: Data to sign
        :param algorithm: Hash algorithm, if None the hash length is determined from ECC curve size
        :return: Signature, r and s coordinates as bytes
        """
        key = private_key if isinstance(private_key, ECC.EccKey) else ECC.import_key(private_key)
        hash_name = algorithm or f"sha{key.pointQ.size_in_bits()}"
        hasher = self._get_algorithm(name=hash_name, data=data)
        signer = DSS.new(key, mode="deterministic-rfc6979")
        return signer.sign(hasher)

    def ecc_verify(
        self,
        public_key: Union[ECC.EccKey, bytes],
        signature: bytes,
        data: bytes,
        algorithm: Optional[str] = None,
    ) -> bool:
        """Verify (EC)DSA signature.

        :param public_key: ECC private or public key, either as EccKey or bytes
        :param signature: Signature to verify, r and s coordinates as bytes
        :param data: Data to validate
        :param algorithm: Hash algorithm, if None the hash length is determined from ECC curve size
        :return: True if the signature is valid
        :raises SPSDKError: Signature length is invalid
        """
        public_key = (
            public_key if isinstance(public_key, ECC.EccKey) else ECC.import_key(public_key)
        )
        hash_name = algorithm or f"sha{public_key.pointQ.size_in_bits()}"
        coordinate_size = public_key.pointQ.size_in_bytes()
        if len(signature) != 2 * coordinate_size:
            raise SPSDKError(
                f"Invalid signature size: expected {2 * coordinate_size}, actual: {len(signature)}"
            )
        hasher = self._get_algorithm(name=hash_name, data=data)
        try:
            DSS.new(public_key, mode="deterministic-rfc6979").verify(hasher, signature)
            return True
        except ValueError:
            return False

    @staticmethod
    def sign_size(key: PrivateKey) -> int:
        """Get size of signature for loaded private key.

        :param key: Private key used to sign data.
        :return: Size of signature in bytes for the private key.
        :raises SPSDKError: Invalid key type.
        """
        if isinstance(key, ECC.EccKey):
            return math.ceil(key.key_size / 8) * 2

        if isinstance(key, RSA.RsaKey):
            return key.key_size // 8

        raise SPSDKError(f"Unsupported private key type to get signature size. {type(key)}")


########################################################################################################################
# SPSDK Backend instance
########################################################################################################################
internal_backend = Backend()  # pylint: disable=invalid-name
